@@ -12,17 +12,6 @@
 #
 # You should have received a copy of the GNU Affero General Public License
 # along with this program.  If not, see <https://www.gnu.org/licenses/>.
-"""A script to generate the SQL needed for lisien's database backend,
-and output it in JSON form.
-
-This uses sqlalchemy to describe the queries. It extends the module of
-the same name in the ``allegedb`` package. If you change anything here,
-you won't be able to use your changes until you put the generated JSON
-where lisien will look for it, as in:
-
-``python3 sqlalchemy.py >sqlite.json``
-
-"""
 
 from collections import OrderedDict
 from functools import partial
@@ -90,6 +79,14 @@ def tables_for_meta(meta):
 		Column("plan_end_tick", INT),
 		sqlite_with_rowid=False,
 	)
+	Table(
+		"bookmarks",
+		meta,
+		Column("key", TEXT, primary_key=True),
+		Column("branch", TEXT, default="trunk"),
+		Column("turn", INT),
+		Column("tick", INT),
+	)
 	Table(
 		"graphs",
 		meta,
@@ -195,7 +192,6 @@ def tables_for_meta(meta):
 		Column("graph", BLOB, primary_key=True),
 		Column("orig", BLOB, primary_key=True),
 		Column("dest", BLOB, primary_key=True),
-		Column("idx", INT, primary_key=True),
 		Column(
 			"branch",
 			TEXT,
@@ -214,7 +210,6 @@ def tables_for_meta(meta):
 		Column("graph", BLOB, primary_key=True),
 		Column("orig", BLOB, primary_key=True),
 		Column("dest", BLOB, primary_key=True),
-		Column("idx", INT, primary_key=True),
 		Column("key", BLOB, primary_key=True),
 		Column(
 			"branch",
@@ -530,9 +525,9 @@ def tables_for_meta(meta):
 		"character_thing_rules_handled",
 		meta,
 		Column("character", BLOB),
-		Column("thing", BLOB),
 		Column("rulebook", BLOB),
 		Column("rule", TEXT),
+		Column("thing", BLOB),
 		Column("branch", TEXT, default="trunk"),
 		Column("turn", INT),
 		Column("tick", INT),
@@ -644,7 +639,6 @@ def queries(table):
 				table["edge_val"].c.graph == bindparam("graph"),
 				table["edge_val"].c.orig == bindparam("orig"),
 				table["edge_val"].c.dest == bindparam("dest"),
-				table["edge_val"].c.idx == bindparam("idx"),
 				table["edge_val"].c.key == bindparam("key"),
 				table["edge_val"].c.branch == bindparam("branch"),
 				or_(
@@ -666,7 +660,6 @@ def queries(table):
 				table["edges"].c.graph == bindparam("graph"),
 				table["edges"].c.orig == bindparam("orig"),
 				table["edges"].c.dest == bindparam("dest"),
-				table["edges"].c.idx == bindparam("idx"),
 				table["edges"].c.branch == bindparam("branch"),
 				or_(
 					table["edges"].c.turn > bindparam("turn"),
@@ -858,7 +851,6 @@ def queries(table):
 			table["edges"].c.graph,
 			table["edges"].c.orig,
 			table["edges"].c.dest,
-			table["edges"].c.idx,
 			table["edges"].c.turn,
 			table["edges"].c.tick,
 			table["edges"].c.extant,
@@ -867,7 +859,6 @@ def queries(table):
 			table["edges"].c.graph,
 			table["edges"].c.orig,
 			table["edges"].c.dest,
-			table["edges"].c.idx,
 			table["edges"].c.turn,
 			table["edges"].c.tick,
 			table["edges"].c.extant,
@@ -892,7 +883,6 @@ def queries(table):
 			table["edge_val"].c.graph,
 			table["edge_val"].c.orig,
 			table["edge_val"].c.dest,
-			table["edge_val"].c.idx,
 			table["edge_val"].c.key,
 			table["edge_val"].c.turn,
 			table["edge_val"].c.tick,
@@ -902,7 +892,6 @@ def queries(table):
 			table["edge_val"].c.graph,
 			table["edge_val"].c.orig,
 			table["edge_val"].c.dest,
-			table["edge_val"].c.idx,
 			table["edge_val"].c.key,
 			table["edge_val"].c.turn,
 			table["edge_val"].c.tick,
@@ -1010,6 +999,19 @@ def queries(table):
 			),
 		)
 	)
+	bookmarks = table["bookmarks"]
+	r["update_bookmark"] = (
+		bookmarks.update()
+		.where(bookmarks.c.key == bindparam("key"))
+		.values(
+			branch=bindparam("branch"),
+			turn=bindparam("turn"),
+			tick=bindparam("tick"),
+		)
+	)
+	r["delete_bookmark"] = bookmarks.delete().where(
+		bookmarks.c.key == bindparam("key")
+	)
 
 	def to_end_clause(tab: Table):
 		return and_(
@@ -1114,7 +1116,7 @@ def queries(table):
 				tab.c.turn < bindparam("turn_to"),
 				and_(
 					tab.c.turn == bindparam("turn_to"),
-					tab.c.tick <= bindparam("tick_to"),
+					tab.c.tick < bindparam("tick_to"),
 				),
 			),
 		)
@@ -1187,18 +1189,25 @@ def queries(table):
 	r["load_rule_actions_tick_to_tick"] = actsel.where(
 		generic_tick_to_tick_clause(act)
 	)
+	kf = table["keyframes"]
+
+	def time_clause(tab):
+		return and_(
+			tab.c.branch == bindparam("branch"),
+			tab.c.turn == bindparam("turn"),
+			tab.c.tick == bindparam("tick"),
+		)
+
+	r["delete_from_keyframes"] = kf.delete().where(time_clause(kf))
+	kfg = table["keyframes_graphs"]
+	r["delete_from_keyframes_graphs"] = kfg.delete().where(time_clause(kfg))
 	kfx = table["keyframe_extensions"]
+	r["delete_from_keyframe_extensions"] = kfx.delete().where(time_clause(kfx))
 	r["get_keyframe_extensions"] = select(
 		kfx.c.universal,
 		kfx.c.rule,
 		kfx.c.rulebook,
-	).where(
-		and_(
-			kfx.c.branch == bindparam("branch"),
-			kfx.c.turn == bindparam("turn"),
-			kfx.c.tick == bindparam("tick"),
-		)
-	)
+	).where(time_clause(kfx))
 
 	for handledtab in (
 		"character_rules_handled",
