@@ -32,16 +32,19 @@ and their node in the physical world is a unit of it.
 
 """
 
+from __future__ import annotations
+
 from abc import abstractmethod
 from collections.abc import Mapping
 from itertools import chain
 from types import MethodType
+from typing import TYPE_CHECKING, Iterable, Iterator
 
 import networkx as nx
 from blinker import Signal
 
 from .cache import FuturistWindowDict, PickyDefaultDict
-from .exc import WorldIntegrityError
+from .exc import AmbiguousLeaderError, WorldIntegrityError
 from .facade import CharacterFacade
 from .graph import (
 	DiGraphPredecessorsMapping,
@@ -51,16 +54,23 @@ from .graph import (
 from .node import Node, Place, Thing
 from .portal import Portal
 from .query import CharacterStatAlias, UnitsAlias
+from .rule import RuleBook
 from .rule import RuleFollower as BaseRuleFollower
 from .rule import RuleMapping
-from .util import (
-	AbstractCharacter,
-	AbstractEngine,
-	getatt,
-	singleton_get,
-	timer,
+from .types import (
+	CharName,
+	EdgeValDict,
+	Key,
+	NodeName,
+	RulebookTypeStr,
+	Stat,
+	StatDict,
 )
-from .wrap import MutableMappingUnwrapper
+from .util import AbstractCharacter, getatt, singleton_get, timer
+from .wrap import MutableMappingUnwrapper, SpecialMapping
+
+if TYPE_CHECKING:
+	from .engine import Engine
 
 
 def grid_2d_8graph(m, n):
@@ -107,7 +117,12 @@ class CharRuleMapping(RuleMapping):
 
 	"""
 
-	def __init__(self, character, rulebook, booktyp):
+	def __init__(
+		self,
+		character: Character,
+		rulebook: RuleBook,
+		booktyp: RulebookTypeStr,
+	):
 		"""Initialize as usual for the ``rulebook``, mostly.
 
 		My ``character`` property will be the one passed in, and my
@@ -122,9 +137,9 @@ class CharRuleMapping(RuleMapping):
 class RuleFollower(BaseRuleFollower):
 	"""Mixin class. Has a rulebook, which you can get a RuleMapping into."""
 
-	character: AbstractCharacter
-	engine: AbstractEngine
-	_book: str
+	character: Character
+	engine: "Engine"
+	_book: RulebookTypeStr
 
 	def _get_rule_mapping(self):
 		return CharRuleMapping(self.character, self.rulebook, self._book)
@@ -157,7 +172,7 @@ class RuleFollower(BaseRuleFollower):
 			self.character.name, branch, turn, tick, n
 		)
 
-	def __contains__(self, k):
+	def __contains__(self, k: Key) -> bool:
 		return self.engine._active_rules_cache.contains_key(
 			self._get_rulebook_name(), *self.engine._btt()
 		)
@@ -205,7 +220,7 @@ class Character(AbstractCharacter, RuleFollower):
 
 	_book = "character"
 
-	def remove_portal(self, origin, destination):
+	def remove_portal(self, origin: NodeName, destination: NodeName):
 		__doc__ = self.remove_edge.__doc__
 		super().remove_edge(origin, destination)
 
@@ -230,7 +245,9 @@ class Character(AbstractCharacter, RuleFollower):
 	def __repr__(self):
 		return "{}.character[{}]".format(repr(self.engine), repr(self.name))
 
-	def __init__(self, engine, name, *, init_rulebooks=True):
+	def __init__(
+		self, engine: "Engine", name: CharName, *, init_rulebooks: bool = True
+	):
 		super().__init__(engine, name)
 		self._avatars_cache = PickyDefaultDict(FuturistWindowDict)
 		if not init_rulebooks:
@@ -250,53 +267,47 @@ class Character(AbstractCharacter, RuleFollower):
 			)
 			cache.store(name, branch, turn, tick, rulebook_name)
 
-	class ThingMapping(MutableMappingUnwrapper, RuleFollower, Signal):
+	class ThingMapping(
+		MutableMappingUnwrapper, SpecialMapping, RuleFollower, Signal
+	):
 		""":class:`Thing` objects that are in a :class:`Character`"""
 
 		_book = "character_thing"
 
-		engine = getatt("character.engine")
-		name = getatt("character.name")
+		engine: "Engine" = getatt("character.engine")
+		name: CharName = getatt("character.name")
 
 		def _get_rulebook_cache(self):
 			return self.engine._characters_things_rulebooks_cache
 
-		def __init__(self, character):
+		def __init__(self, character: Character):
 			"""Store the character and initialize cache."""
-			super().__init__()
+			super().__init__(character)
 			self.character = character
 
 		def __iter__(self):
 			cache = self.engine._things_cache
 			char = self.name
 			branch, turn, tick = self.engine._btt()
-			for key in cache.iter_keys(char, branch, turn, tick):
-				try:
-					if (
-						cache.retrieve(char, key, branch, turn, tick)
-						is not None
-					):
-						yield key
-				except KeyError:
-					continue
+			return cache.iter_things(char, branch, turn, tick)
 
 		def __contains__(self, thing):
 			branch, turn, tick = self.engine._btt()
 			args = self.character.name, thing, branch, turn, tick
 			cache = self.engine._things_cache
-			return cache.contains_key(*args)
+			return cache.thing_exists(*args)
 
 		def __len__(self):
 			return self.engine._things_cache.count_keys(
 				self.character.name, *self.engine._btt()
 			)
 
-		def __getitem__(self, thing):
+		def __getitem__(self, thing: NodeName):
 			if thing not in self:
 				raise KeyError("No such thing: {}".format(thing))
 			return self._make_thing(thing)
 
-		def _make_thing(self, thing, val=None):
+		def _make_thing(self, thing: NodeName, val: Thing | None = None):
 			cache = self.engine._node_objs
 			if isinstance(val, Thing):
 				th = cache[self.name, thing] = val
@@ -308,7 +319,7 @@ class Character(AbstractCharacter, RuleFollower):
 				th = cache[(self.name, thing)] = Thing(self.character, thing)
 			return th
 
-		def __setitem__(self, thing, val: Mapping):
+		def __setitem__(self, thing: NodeName, val: Mapping):
 			if not isinstance(val, Mapping):
 				raise TypeError("Things are made from Mappings")
 			if "location" not in val:
@@ -332,11 +343,11 @@ class Character(AbstractCharacter, RuleFollower):
 				tick,
 				val.pop("location"),
 			)
-			th = self._make_thing(thing, val)
+			th = self._make_thing(thing)
 			th.clear()
 			th.update({k: v for (k, v) in val.items() if k != "location"})
 
-		def __delitem__(self, thing):
+		def __delitem__(self, thing: NodeName):
 			self[thing].delete()
 
 		def __repr__(self):
@@ -344,7 +355,9 @@ class Character(AbstractCharacter, RuleFollower):
 				repr(self.engine), repr(self.name)
 			)
 
-	class PlaceMapping(MutableMappingUnwrapper, RuleFollower, Signal):
+	class PlaceMapping(
+		MutableMappingUnwrapper, SpecialMapping, RuleFollower, Signal
+	):
 		""":class:`Place` objects that are in a :class:`Character`"""
 
 		_book = "character_place"
@@ -355,9 +368,9 @@ class Character(AbstractCharacter, RuleFollower):
 		def update(self, __m: dict, **kwargs) -> None:
 			self.character.node.update(__m, **kwargs)
 
-		def __init__(self, character):
+		def __init__(self, character: Character):
 			"""Store the character."""
-			super().__init__()
+			super().__init__(character)
 			self.character = character
 			self.engine = engine = character.engine
 			charn = character.name
@@ -409,14 +422,14 @@ class Character(AbstractCharacter, RuleFollower):
 				pass
 			return n
 
-		def __contains__(self, place):
+		def __contains__(self, place: NodeName) -> bool:
 			nodes_contains, things_contains, charn, btt = self._contains_stuff
 			branch, turn, tick = btt()
 			return nodes_contains(
 				charn, place, branch, turn, tick
 			) and not things_contains(charn, place, branch, turn, tick)
 
-		def __getitem__(self, place):
+		def __getitem__(self, place: NodeName) -> Place:
 			(nodes_contains, things_contains, charn, btt, cache, character) = (
 				self._get_stuff
 			)
@@ -432,7 +445,7 @@ class Character(AbstractCharacter, RuleFollower):
 				return ret
 			return cache[(charn, place)]
 
-		def __setitem__(self, place, v):
+		def __setitem__(self, place: NodeName, v: Place | StatDict):
 			(node_exists, exist_node, get_node, charn, character) = (
 				self._set_stuff
 			)
@@ -444,7 +457,7 @@ class Character(AbstractCharacter, RuleFollower):
 				)
 			pl.update(v)
 
-		def __delitem__(self, place):
+		def __delitem__(self, place: NodeName):
 			self[place].delete()
 
 		def __repr__(self):
@@ -452,16 +465,14 @@ class Character(AbstractCharacter, RuleFollower):
 				repr(self.character.engine), repr(self.character.name)
 			)
 
-	class ThingPlaceMapping(GraphNodeMapping, Signal):
+	class ThingPlaceMapping(GraphNodeMapping, SpecialMapping, Signal):
 		"""GraphNodeMapping but for Place and Thing"""
 
-		_book = "character_node"
+		character: Character = getatt("graph")
+		engine: "Engine" = getatt("db")
+		name: CharName = getatt("character.name")
 
-		character = getatt("graph")
-		engine = getatt("db")
-		name = getatt("character.name")
-
-		def __init__(self, character):
+		def __init__(self, character: Character):
 			"""Store the character."""
 			super().__init__(character)
 			Signal.__init__(self)
@@ -482,20 +493,20 @@ class Character(AbstractCharacter, RuleFollower):
 			)
 			self._placemap = character.place
 
-		def __contains__(self, k):
+		def __contains__(self, k: NodeName):
 			node_exists, charn = self._contains_stuff
 			return node_exists(charn, k)
 
-		def __getitem__(self, k):
+		def __getitem__(self, k: NodeName) -> Thing | Place:
 			node_exists, charn, get_node, character = self._getitem_stuff
 			if not node_exists(charn, k):
 				raise KeyError("No such node: " + str(k))
 			return get_node(character, k)
 
-		def __setitem__(self, k, v):
+		def __setitem__(self, k: NodeName, v: Place | StatDict):
 			self._placemap[k] = v
 
-		def __delitem__(self, k):
+		def __delitem__(self, k: NodeName):
 			(node_exists, charn, is_thing, thingmap, placemap) = (
 				self._delitem_stuff
 			)
@@ -508,7 +519,9 @@ class Character(AbstractCharacter, RuleFollower):
 
 	node_map_cls = ThingPlaceMapping
 
-	class PortalSuccessorsMapping(DiGraphSuccessorsMapping, RuleFollower):
+	class PortalSuccessorsMapping(
+		DiGraphSuccessorsMapping, SpecialMapping, RuleFollower
+	):
 		"""Mapping of nodes that have at least one outgoing edge.
 
 		Maps them to another mapping, keyed by the destination nodes,
@@ -518,8 +531,8 @@ class Character(AbstractCharacter, RuleFollower):
 
 		_book = "character_portal"
 
-		character = getatt("graph")
-		engine = getatt("graph.engine")
+		character: Character = getatt("graph")
+		engine: "Engine" = getatt("graph.engine")
 
 		def __init__(self, graph):
 			super().__init__(graph)
@@ -532,7 +545,7 @@ class Character(AbstractCharacter, RuleFollower):
 		def _get_rulebook_cache(self):
 			return self._cporh
 
-		def __getitem__(self, orig):
+		def __getitem__(self, orig: NodeName) -> Successors:
 			node_exists, charn, cache = self._getitem_stuff
 			if node_exists(charn, orig):
 				if orig not in cache:
@@ -543,7 +556,7 @@ class Character(AbstractCharacter, RuleFollower):
 		def __delitem__(self, orig):
 			super().__delitem__(orig)
 
-		def update(self, other, **kwargs):
+		def update(self, other: EdgeValDict, **kwargs):
 			"""Recursively update the stats of all portals
 
 			Input should be a dictionary of dictionaries of dictionaries
@@ -556,6 +569,7 @@ class Character(AbstractCharacter, RuleFollower):
 			delete them.
 
 			"""
+			kwargs: EdgeValDict
 			engine = self.engine
 			planning = engine._planning
 			forward = engine._forward
@@ -572,12 +586,11 @@ class Character(AbstractCharacter, RuleFollower):
 			):
 				for orig, dests in chain(other.items(), kwargs.items()):
 					for dest, kvs in dests.items():
-						if kvs is None:
+						if kvs is ...:
 							for k in iter_edge_keys(
 								charn,
 								orig,
 								dest,
-								0,
 								branch,
 								turn,
 								start_tick,
@@ -587,12 +600,11 @@ class Character(AbstractCharacter, RuleFollower):
 									charn,
 									orig,
 									dest,
-									0,
 									k,
 									branch,
 									turn,
 									tick,
-									None,
+									...,
 									planning=planning,
 									forward=forward,
 									loading=True,
@@ -601,19 +613,17 @@ class Character(AbstractCharacter, RuleFollower):
 									charn,
 									orig,
 									dest,
-									0,
 									k,
 									branch,
 									turn,
 									tick,
-									None,
+									...,
 								)
 								tick += 1
 							store_edge(
 								charn,
 								orig,
 								dest,
-								0,
 								branch,
 								turn,
 								tick,
@@ -623,7 +633,7 @@ class Character(AbstractCharacter, RuleFollower):
 								loading=True,
 							)
 							exist_edge(
-								charn, orig, dest, 0, branch, turn, tick, False
+								charn, orig, dest, branch, turn, tick, False
 							)
 							tick += 1
 						else:
@@ -631,7 +641,6 @@ class Character(AbstractCharacter, RuleFollower):
 								charn,
 								orig,
 								dest,
-								0,
 								branch,
 								turn,
 								tick,
@@ -641,7 +650,7 @@ class Character(AbstractCharacter, RuleFollower):
 								loading=True,
 							)
 							exist_edge(
-								charn, orig, dest, 0, branch, turn, tick, True
+								charn, orig, dest, branch, turn, tick, True
 							)
 							tick += 1
 							for k, v in kvs.items():
@@ -649,7 +658,6 @@ class Character(AbstractCharacter, RuleFollower):
 									charn,
 									orig,
 									dest,
-									0,
 									k,
 									branch,
 									turn,
@@ -663,7 +671,6 @@ class Character(AbstractCharacter, RuleFollower):
 									charn,
 									orig,
 									dest,
-									0,
 									k,
 									branch,
 									turn,
@@ -676,9 +683,13 @@ class Character(AbstractCharacter, RuleFollower):
 		class Successors(DiGraphSuccessorsMapping.Successors):
 			"""Mapping for possible destinations from some node."""
 
-			engine = getatt("graph.engine")
+			engine: "Engine" = getatt("graph.engine")
 
-			def __init__(self, container, orig):
+			def __init__(
+				self,
+				container: Character.PortalSuccessorsMapping,
+				orig: NodeName,
+			):
 				super().__init__(container, orig)
 				graph = self.graph
 				engine = graph.engine
@@ -695,14 +706,16 @@ class Character(AbstractCharacter, RuleFollower):
 					engine._nbtt,
 				)
 
-			def __getitem__(self, dest):
+			def __getitem__(self, dest: NodeName) -> Portal:
 				get_edge, graph, orig = self._getitem_stuff
 				if dest in self:
-					return get_edge(graph, orig, dest, 0)
-				raise KeyError("No such portal: {}->{}".format(orig, dest))
+					return get_edge(graph, orig, dest)
+				raise KeyError("No such portal", graph, orig, dest)
 
-			def __setitem__(self, dest, value):
-				if value is None:
+			def __setitem__(
+				self, dest: NodeName, value: Portal | StatDict | ...
+			):
+				if value is ...:
 					del self[dest]
 					return
 				(
@@ -720,19 +733,22 @@ class Character(AbstractCharacter, RuleFollower):
 				for k, v in value.items():
 					branch, turn, tick = nbtt()
 					db_edge_val_set(
-						charn, orig, dest, 0, k, branch, turn, tick, v
+						charn, orig, dest, k, branch, turn, tick, v
 					)
 					edge_val_cache_store(
-						charn, orig, dest, 0, k, branch, turn, tick, v
+						charn, orig, dest, k, branch, turn, tick, v
 					)
 
-			def __delitem__(self, dest):
+			def __delitem__(self, dest: NodeName):
 				if dest not in self:
 					raise KeyError("No portal to {}".format(dest))
 				self[dest].delete()
 
-			def update(self, other=None, **kwargs):
-				if other is None:
+			def update(
+				self, other: dict[NodeName, StatDict] | None = None, **kwargs
+			):
+				kwargs: dict[NodeName, StatDict]
+				if other is ...:
 					it = kwargs.items()
 				else:
 					it = chain(other.items(), kwargs.items())
@@ -741,7 +757,9 @@ class Character(AbstractCharacter, RuleFollower):
 
 	adj_cls = PortalSuccessorsMapping
 
-	class PortalPredecessorsMapping(DiGraphPredecessorsMapping, RuleFollower):
+	class PortalPredecessorsMapping(
+		DiGraphPredecessorsMapping, SpecialMapping, RuleFollower
+	):
 		"""Mapping of nodes that have at least one incoming edge.
 
 		Maps to another mapping keyed by the origin nodes, which maps to
@@ -751,7 +769,7 @@ class Character(AbstractCharacter, RuleFollower):
 
 		_book = "character_portal"
 
-		def __init__(self, graph):
+		def __init__(self, graph: Character):
 			super().__init__(graph)
 			self._cporc = graph.engine._characters_portals_rulebooks_cache
 
@@ -761,7 +779,11 @@ class Character(AbstractCharacter, RuleFollower):
 		class Predecessors(DiGraphPredecessorsMapping.Predecessors):
 			"""Mapping of possible origins from some destination."""
 
-			def __init__(self, container, dest):
+			def __init__(
+				self,
+				container: Character.PortalPredecessorsMapping,
+				dest: NodeName,
+			):
 				super().__init__(container, dest)
 				graph = self.graph
 				self._setitem_stuff = (
@@ -771,7 +793,7 @@ class Character(AbstractCharacter, RuleFollower):
 					self.db._edge_objs,
 				)
 
-			def __setitem__(self, orig, value):
+			def __setitem__(self, orig: NodeName, value: StatDict):
 				graph, graph_name, dest, portal_objs = self._setitem_stuff
 				key = (graph_name, orig, dest)
 				if key not in portal_objs:
@@ -783,7 +805,7 @@ class Character(AbstractCharacter, RuleFollower):
 
 	pred_cls = PortalPredecessorsMapping
 
-	class UnitGraphMapping(Mapping, RuleFollower):
+	class UnitGraphMapping(SpecialMapping, RuleFollower):
 		"""A mapping of other characters in which one has a unit."""
 
 		_book = "unit"
@@ -794,7 +816,7 @@ class Character(AbstractCharacter, RuleFollower):
 		def _get_rulebook_cache(self):
 			return self._avrc
 
-		def __init__(self, char):
+		def __init__(self, char: Character):
 			"""Remember my character."""
 			self.character = char
 			self._char_av_cache = {}
@@ -806,9 +828,9 @@ class Character(AbstractCharacter, RuleFollower):
 			charn = char.name
 			btt = engine._btt
 			self._iter_stuff = (get_char_graphs, charn, btt)
-			self._len_stuff = (avcache.count_entities_or_keys, charn, btt)
+			self._len_stuff = (avcache.count_graphs, charn, btt)
 			self._contains_stuff = (
-				avcache.contains_key,
+				avcache.dict_cache.contains_key,
 				charn,
 				btt,
 			)
@@ -826,15 +848,15 @@ class Character(AbstractCharacter, RuleFollower):
 				engine.character,
 			)
 
-		def __call__(self, av):
+		def __call__(self, unit: Place | Thing):
 			"""Add the unit
 
 			It must be an instance of Place or Thing.
 
 			"""
-			if av.__class__ not in (Place, Thing):
+			if unit.__class__ not in (Place, Thing):
 				raise TypeError("Only Things and Places may be units")
-			self._add_av(av.name, av.character.name)
+			self._add_av(unit.name, unit.character.name)
 
 		def __iter__(self):
 			"""Iterate over graphs with unit nodes in them"""
@@ -843,10 +865,10 @@ class Character(AbstractCharacter, RuleFollower):
 				if graph in self:
 					yield graph
 
-		def __contains__(self, k):
+		def __contains__(self, k: NodeName):
 			retrieve, charn, btt = self._contains_stuff
 			got = retrieve(charn, k, *btt())
-			return got is not None and not isinstance(got, Exception)
+			return got is not ... and not isinstance(got, Exception)
 
 		def __len__(self):
 			"""Number of graphs in which I have a unit."""
@@ -855,7 +877,7 @@ class Character(AbstractCharacter, RuleFollower):
 				pass
 			return n
 
-		def _get_char_av_cache(self, g):
+		def _get_char_av_cache(self, g: CharName):
 			if g not in self:
 				raise KeyError
 			if g not in self._char_av_cache:
@@ -866,7 +888,7 @@ class Character(AbstractCharacter, RuleFollower):
 			return self._get_char_av_cache(g)
 
 		@property
-		def node(self):
+		def node(self) -> dict[NodeName, Place | Thing]:
 			"""If I have units in only one graph, return a map of them
 
 			Otherwise, raise AttributeError.
@@ -884,7 +906,7 @@ class Character(AbstractCharacter, RuleFollower):
 				)
 
 		@property
-		def only(self):
+		def only(self) -> Place | Thing:
 			"""If I have only one unit, this is it
 
 			Otherwise, raise AttributeError.
@@ -900,23 +922,25 @@ class Character(AbstractCharacter, RuleFollower):
 		class CharacterUnitMapping(Mapping):
 			"""Mapping of units of one Character in another Character."""
 
-			def __init__(self, outer, graphn):
+			def __init__(
+				self, outer: Character.UnitGraphMapping, graphn: CharName
+			):
 				"""Store this character and the name of the other one"""
-				self.character = character = outer.character
+				self.character = outer.character
 				self.engine = engine = outer.engine
 				self.name = name = outer.name
 				self.graph = graphn
-				avcache = engine._unitness_cache
+				unitcache = engine._unitness_cache
 				btt = engine._btt
 				self._iter_stuff = iter_stuff = (
-					avcache.get_char_graph_units,
-					avcache.contains_key,
+					unitcache.get_char_graph_units,
+					unitcache.contains_unit,
 					name,
 					graphn,
 					btt,
 				)
 				self._contains_stuff = (
-					avcache._base_retrieve,
+					unitcache._base_retrieve,
 					name,
 					graphn,
 					btt,
@@ -927,7 +951,12 @@ class Character(AbstractCharacter, RuleFollower):
 					graphn,
 					engine.character,
 				)
-				self._only_stuff = (get_node, engine.character, graphn)
+				self._only_stuff = (
+					get_node,
+					engine.character,
+					graphn,
+					engine._btt,
+				)
 
 			def __iter__(self):
 				"""Iterate over names of unit nodes"""
@@ -938,7 +967,7 @@ class Character(AbstractCharacter, RuleFollower):
 					if validate(name, graphn, unit, *btt()):
 						yield unit
 
-			def __contains__(self, av):
+			def __contains__(self, av: NodeName):
 				base_retrieve, name, graphn, btt = self._contains_stuff
 				return (
 					base_retrieve(
@@ -960,7 +989,7 @@ class Character(AbstractCharacter, RuleFollower):
 						n += 1
 				return n
 
-			def __getitem__(self, av):
+			def __getitem__(self, av: NodeName) -> Place | Thing:
 				(
 					get_char_graph_avs,
 					unitness_cache_has,
@@ -976,20 +1005,33 @@ class Character(AbstractCharacter, RuleFollower):
 				raise KeyError("No unit: {}".format(av))
 
 			@property
-			def only(self):
+			def only(self) -> Place | Thing:
 				"""If I have only one unit, return it; else error"""
+
+				get_node, charmap, graphn, btt = self._only_stuff
 				mykey = singleton_get(self.keys())
 				if mykey is None:
-					raise AttributeError("No unit, or more than one")
-				get_node, charmap, graphn = self._only_stuff
-				return get_node(charmap[graphn], mykey)
+					raise AttributeError(
+						"No unit, or more than one",
+						self.character.name,
+						*self.engine._btt(),
+					)
+				try:
+					return get_node(charmap[graphn], mykey)
+				except KeyError as ex:
+					raise AttributeError(
+						"The unit doesn't exist",
+						graphn,
+						mykey,
+						*self.engine._btt(),
+					) from ex
 
 			def __repr__(self):
 				return "{}.character[{}].unit".format(
 					repr(self.engine), repr(self.name)
 				)
 
-	def facade(self):
+	def facade(self) -> CharacterFacade:
 		"""Return a temporary copy of this Character
 
 		A Facade looks like its :class:`Character`, but doesn't do any of the
@@ -999,35 +1041,38 @@ class Character(AbstractCharacter, RuleFollower):
 		"""
 		return CharacterFacade(self)
 
-	def add_place(self, node_for_adding, **attr):
+	def add_place(self, node_for_adding: NodeName, **attr):
 		"""Add a new Place"""
+		attr: StatDict
 		self.place[node_for_adding] = attr
 
-	def add_places_from(self, seq, **attrs):
+	def add_places_from(self, seq: Iterable[NodeName], **attrs):
+		attrs: StatDict
 		for place in seq:
 			self.add_place(place, **attrs)
 
-	def remove_place(self, place):
+	def remove_place(self, place: NodeName) -> None:
 		"""Remove an existing Place"""
 		if place in self.place:
 			self.remove_node(place)
 		else:
 			raise KeyError("No such place: {}".format(place))
 
-	def remove_thing(self, thing):
+	def remove_thing(self, thing: NodeName) -> None:
 		"""Remove an existing Thing"""
 		if thing in self.thing:
 			self.remove_node(thing)
 		else:
 			raise KeyError("No such thing: {}".format(thing))
 
-	def add_thing(self, name, location, **kwargs):
+	def add_thing(self, name: NodeName, location: NodeName, **kwargs):
 		"""Make a new Thing and set its location"""
+		kwargs: StatDict
 		if name in self.thing:
 			raise WorldIntegrityError(
 				"Already have a Thing named {}".format(name)
 			)
-		starter = self.node_dict_factory()
+		starter = StatDict(self.node_dict_factory())
 		starter.update(kwargs)
 		if isinstance(location, Node):
 			location = location.name
@@ -1037,15 +1082,22 @@ class Character(AbstractCharacter, RuleFollower):
 			self._succ[name] = self.adjlist_inner_dict_factory()
 			self._pred[name] = self.adjlist_inner_dict_factory()
 
-	def add_things_from(self, seq, **attrs):
+	def add_things_from(
+		self,
+		seq: Iterable[
+			tuple[NodeName, NodeName] | tuple[NodeName, NodeName, StatDict]
+		],
+		**attrs,
+	) -> None:
 		"""Make many new Things"""
+		attrs: StatDict
 		for tup in seq:
 			name = tup[0]
 			location = tup[1]
 			kwargs = tup[2] if len(tup) > 2 else attrs
 			self.add_thing(name, location, **kwargs)
 
-	def place2thing(self, name, location):
+	def place2thing(self, name: NodeName, location: NodeName) -> None:
 		"""Turn a Place into a Thing with the given location.
 
 		It will keep all its attached Portals.
@@ -1061,7 +1113,7 @@ class Character(AbstractCharacter, RuleFollower):
 				port.destination = thing
 			self.engine._node_objs[self.name, name] = thing
 
-	def thing2place(self, name):
+	def thing2place(self, name: NodeName) -> None:
 		"""Unset a Thing's location, and thus turn it into a Place."""
 		self.engine._set_thing_loc(self.name, name, None)
 		if (self.name, name) in self.engine._node_objs:
@@ -1073,12 +1125,15 @@ class Character(AbstractCharacter, RuleFollower):
 				port.destination = place
 			self.engine._node_objs[self.name, name] = place
 
-	def add_portal(self, origin, destination, **kwargs):
+	def add_portal(
+		self, origin: NodeName, destination: NodeName, **kwargs
+	) -> None:
 		"""Connect the origin to the destination with a :class:`Portal`.
 
 		Keyword arguments are attributes of the :class:`Portal`.
 
 		"""
+		kwargs: StatDict
 		if isinstance(origin, Node):
 			origin = origin.name
 		if origin not in self.place and origin not in self.thing:
@@ -1089,24 +1144,27 @@ class Character(AbstractCharacter, RuleFollower):
 			self.add_place(destination)
 		branch, turn, tick = self.engine._nbtt()
 		self.engine._edges_cache.store(
-			self.name, origin, destination, 0, branch, turn, tick, True
+			self.name, origin, destination, branch, turn, tick, True
 		)
 		self.engine.query.exist_edge(
-			self.name, origin, destination, 0, branch, turn, tick, True
+			self.name, origin, destination, branch, turn, tick, True
 		)
 		for k, v in kwargs.items():
 			branch, turn, tick = self.engine._nbtt()
 			self.engine._edge_val_cache.store(
-				self.name, origin, destination, 0, k, branch, turn, tick, v
+				self.name, origin, destination, k, branch, turn, tick, v
 			)
 			self.engine.query.edge_val_set(
-				self.name, origin, destination, 0, k, branch, turn, tick, v
+				self.name, origin, destination, k, branch, turn, tick, v
 			)
 
-	def new_portal(self, origin, destination, **kwargs):
+	def new_portal(
+		self, origin: NodeName, destination: NodeName, **kwargs
+	) -> Portal:
 		"""Create a portal and return it"""
+		kwargs: StatDict
 		self.add_portal(origin, destination, **kwargs)
-		return self.engine._get_edge(self, origin, destination, 0)
+		return self.engine._get_edge(self, origin, destination)
 
 	def add_portals_from(self, seq, **kwargs):
 		"""Make portals for a sequence of (origin, destination) pairs
@@ -1120,7 +1178,11 @@ class Character(AbstractCharacter, RuleFollower):
 			kwarrgs = tup[2] if len(tup) > 2 else kwargs
 			self.add_portal(orig, dest, **kwarrgs)
 
-	def add_unit(self, a, b=None):
+	def add_unit(
+		self,
+		a: Place | Thing | Character | CharName,
+		b: Place | Thing | NodeName | None = None,
+	) -> None:
 		"""Start keeping track of a unit.
 
 		Units are nodes in other characters that are in some sense part of
@@ -1147,7 +1209,7 @@ class Character(AbstractCharacter, RuleFollower):
 		else:
 			if isinstance(a, Character):
 				g = a.name
-			elif not isinstance(a, str):
+			elif not isinstance(a, Key):
 				raise TypeError(
 					"when called with two arguments, "
 					"the first is a character or its name"
@@ -1156,7 +1218,7 @@ class Character(AbstractCharacter, RuleFollower):
 				g = a
 			if isinstance(b, (Place, Thing)):
 				n = b.name
-			elif not isinstance(b, str):
+			elif not isinstance(b, Key):
 				raise TypeError(
 					"when called with two arguments, "
 					"the second is a thing/place or its name"
@@ -1173,7 +1235,11 @@ class Character(AbstractCharacter, RuleFollower):
 		)
 		self.engine.query.unit_set(self.name, g, n, branch, turn, tick, True)
 
-	def remove_unit(self, a, b=None):
+	def remove_unit(
+		self,
+		a: Place | Thing | Character | CharName,
+		b: Place | Thing | NodeName | None = None,
+	):
 		"""This is no longer my unit, though it still exists"""
 		if self.engine._planning:
 			raise NotImplementedError(
@@ -1197,12 +1263,12 @@ class Character(AbstractCharacter, RuleFollower):
 		)
 		self.engine.query.unit_set(self.name, g, n, branch, turn, tick, False)
 
-	def portals(self):
+	def portals(self) -> Iterator[Portal]:
 		"""Iterate over all portals."""
 		for o in self.adj.values():
 			yield from o.values()
 
-	def historical(self, stat):
+	def historical(self, stat: Stat) -> UnitsAlias | CharacterStatAlias:
 		"""Get a historical view on the given stat
 
 		This functions like the value of the stat, but changes
