@@ -26,12 +26,13 @@ from blinker import Signal
 from .cache import Cache, TurnEndDict, TurnEndPlanDict, UnitnessCache
 from .exc import NotInKeyframeError, TotalKeyError
 from .graph import DiGraph, Edge, Node
-from .typing import CharName, Key, NodeName
+from .types import CharName, Key, NodeName
 from .util import (
 	AbstractCharacter,
 	AbstractEngine,
 	AbstractThing,
 	SignalDict,
+	TimeSignalDescriptor,
 	getatt,
 	timer,
 )
@@ -41,7 +42,7 @@ from .xcollections import CompositeDict
 
 class FacadeEntity(MutableMapping, Signal, ABC):
 	exists = True
-	graph = getatt("character")
+	character = getatt("graph")
 
 	@property
 	def rulebook(self):
@@ -53,18 +54,12 @@ class FacadeEntity(MutableMapping, Signal, ABC):
 	def rulebook(self, rbname):
 		self._patch["rulebook"] = rbname
 
-	@property
-	def engine(self):
-		return self.character.engine
-
-	db = engine
-
 	@abstractmethod
 	def _get_real(self, name):
 		raise NotImplementedError()
 
 	def __init__(self, mapping, real_or_name=None, **kwargs):
-		self.facade = self.character = getattr(mapping, "facade", mapping)
+		self.facade = self.graph = getattr(mapping, "facade", mapping)
 		self._mapping = mapping
 		is_name = not hasattr(real_or_name, "name") and not hasattr(
 			real_or_name, "orig"
@@ -84,7 +79,7 @@ class FacadeEntity(MutableMapping, Signal, ABC):
 	def __contains__(self, item):
 		patch = self._patch
 		if item in patch:
-			return patch[item] is not None
+			return patch[item] is not ...
 		if hasattr(self, "_real"):
 			return item in self._real
 		return False
@@ -95,7 +90,7 @@ class FacadeEntity(MutableMapping, Signal, ABC):
 		if hasattr(self, "_real"):
 			ks |= self._real.keys()
 		for k in ks:
-			if k not in patch or patch[k] is not None:
+			if k not in patch or patch[k] is not ...:
 				yield k
 
 	def __len__(self):
@@ -106,7 +101,7 @@ class FacadeEntity(MutableMapping, Signal, ABC):
 
 	def __getitem__(self, k):
 		if k in self._patch:
-			if self._patch[k] is None:
+			if self._patch[k] is ...:
 				raise KeyError("{} has been masked.".format(k))
 			return self._patch[k]
 		if not hasattr(self, "_real"):
@@ -133,7 +128,7 @@ class FacadeEntity(MutableMapping, Signal, ABC):
 		self._patch[k] = v
 
 	def __delitem__(self, k):
-		self._patch[k] = None
+		self._patch[k] = ...
 
 	def apply(self):
 		self._real.update(self._patch)
@@ -181,13 +176,13 @@ class FacadeEntityMapping(MutableMappingUnwrapper, Signal, ABC):
 
 	def __contains__(self, k):
 		if k in self._patch:
-			return self._patch[k] is not None
+			return self._patch[k] is not ...
 		return k in self._get_inner_map()
 
 	def __iter__(self):
 		seen = set()
 		for k in self._patch:
-			if k not in seen and self._patch[k] is not None:
+			if k not in seen and self._patch[k] is not ...:
 				yield k
 			seen.add(k)
 		for k in self._get_inner_map():
@@ -211,7 +206,7 @@ class FacadeEntityMapping(MutableMappingUnwrapper, Signal, ABC):
 				v = {}
 			self._patch[k] = self._make(k, v)
 		ret = self._patch[k]
-		if ret is None:
+		if ret is ...:
 			raise KeyError(k)
 		if type(ret) is not self.facadecls:
 			ret = self._patch[k] = self._make(k, ret)
@@ -237,7 +232,7 @@ class FacadeEntityMapping(MutableMappingUnwrapper, Signal, ABC):
 			user: CharacterFacade
 			for user in list(that.users()):
 				user.remove_unit(self.facade.name, k)
-		self._patch[k] = None
+		self._patch[k] = ...
 
 
 class FacadeRulebook(MutableSequence, ABC):
@@ -331,14 +326,14 @@ class FacadeNode(FacadeEntity, Node):
 		def __iter__(self):
 			engine = self._entity.engine
 			charn = self._entity.character.name
-			return engine._unitness_cache.user_cache.iter_keys(
+			return engine._unitness_cache.leader_cache.iter_keys(
 				charn, self._entity.name, *engine._btt()
 			)
 
 		def __len__(self):
 			engine = self._entity.engine
 			charn = self._entity.character.name
-			return engine._unitness_cache.user_cache.count_keys(
+			return engine._unitness_cache.leader_cache.count_keys(
 				charn, self._entity.name, *engine._btt()
 			)
 
@@ -347,7 +342,7 @@ class FacadeNode(FacadeEntity, Node):
 			charn = self._entity.character.name
 			try:
 				return bool(
-					engine._unitness_cache.user_cache.retrieve(
+					engine._unitness_cache.leader_cache.retrieve(
 						charn, self._entity.name, item, *engine._btt()
 					)
 				)
@@ -367,7 +362,9 @@ class FacadeNode(FacadeEntity, Node):
 			self._entity = node
 
 		def __iter__(self):
-			if hasattr(self._entity.engine, "_node_contents_cache"):
+			if hasattr(self._entity, "engine") and hasattr(
+				self._entity.engine, "_node_contents_cache"
+			):
 				# The real contents cache is wrapped by the facade engine.
 				try:
 					return self._entity.engine._node_contents_cache.retrieve(
@@ -506,7 +503,7 @@ class FacadePortalMapping(FacadeEntityMapping, ABC):
 		if node not in self._patch:
 			self._patch[node] = self.cls(self.facade, node)
 		ret = self._patch[node]
-		if ret is None:
+		if ret is ...:
 			raise KeyError("masked")
 		if type(ret) is not self.cls:
 			nuret = self.cls(self.facade, node)
@@ -522,16 +519,33 @@ class FacadePortal(FacadeEntity, Edge):
 	"""Lightweight analogue of Portal for Facade use."""
 
 	def __init__(self, mapping, other, **kwargs):
-		super().__init__(mapping, other, **kwargs)
 		if hasattr(mapping, "orig"):
 			self.orig = mapping.orig
 			self.dest = other
 		else:
 			self.dest = mapping.dest
 			self.orig = other
+		if hasattr(mapping, "facade"):
+			facade = mapping.facade
+		else:
+			facade = mapping
 		try:
-			self._real = self.facade.character.portal[self.orig][self.dest]
+			super().__init__(
+				facade.character.node[self.orig],
+				facade.character.node[self.dest],
+				**kwargs,
+			)
+			self._real = facade.character.portal[self.orig][self.dest]
 		except (KeyError, AttributeError):
+			if self.orig in facade.node:
+				origin = facade.node[self.orig]
+			else:
+				origin = facade.new_place(self.orig)
+			if self.dest in facade.node:
+				destination = facade.node[self.dest]
+			else:
+				destination = facade.new_place(self.dest)
+			super().__init__(origin, destination, **kwargs)
 			self._real = {}
 
 	def __getitem__(self, item):
@@ -939,7 +953,7 @@ class CharacterFacade(AbstractCharacter):
 						yield k
 						seen.add(k)
 			for k, v in self._patch.items():
-				if k not in seen and v is not None:
+				if k not in seen and v is not ...:
 					yield k
 
 		def __len__(self):
@@ -950,7 +964,7 @@ class CharacterFacade(AbstractCharacter):
 
 		def __contains__(self, k):
 			if k in self._patch:
-				return self._patch[k] is not None
+				return self._patch[k] is not ...
 			if (
 				hasattr(self.facade.character, "graph")
 				and k in self.facade.character.graph
@@ -966,8 +980,8 @@ class CharacterFacade(AbstractCharacter):
 				if not hasattr(ret, "unwrap"):
 					return ret
 				self._patch[k] = ret.unwrap()
-			if self._patch[k] is None:
-				return KeyError
+			if self._patch[k] is ...:
+				return KeyError("masked", k)
 			return self._patch[k]
 
 		def __setitem__(self, k, v):
@@ -979,7 +993,7 @@ class CharacterFacade(AbstractCharacter):
 			self._patch[k] = v
 
 		def __delitem__(self, k):
-			self._patch[k] = None
+			self._patch[k] = ...
 
 	def apply(self):
 		"""Do all my changes for real in a batch"""
@@ -990,13 +1004,13 @@ class CharacterFacade(AbstractCharacter):
 		realport = realchar.portal
 		realeng = self.engine._real
 		for k, v in self.stat._patch.items():
-			if v is None:
+			if v is ...:
 				del realstat[k]
 			else:
 				realstat[k] = v
 		self.stat._patch = {}
 		for k, v in self.thing._patch.items():
-			if v is None:
+			if v is ...:
 				del realthing[k]
 			elif k not in realthing:
 				if isinstance(v, FacadeThing):
@@ -1008,7 +1022,7 @@ class CharacterFacade(AbstractCharacter):
 				v.apply()
 		self.thing._patch = {}
 		for k, v in self.place._patch.items():
-			if v is None:
+			if v is ...:
 				del realplace[k]
 			elif k not in realplace:
 				realchar.add_place(k, **v)
@@ -1020,7 +1034,7 @@ class CharacterFacade(AbstractCharacter):
 				if not getattr(dests, "_tampered", False):
 					continue
 				for dest, v in dests.items():
-					if v is None:
+					if v is ...:
 						del realport[orig][dest]
 					elif orig not in realport or dest not in realport[orig]:
 						realchar.add_portal(orig, dest, **v)
@@ -1036,18 +1050,22 @@ class EngineFacade(AbstractEngine):
 	thing_cls = FacadeThing
 	place_cls = FacadePlace
 	portal_cls = FacadePortal
+	time = TimeSignalDescriptor()
 
-	class FacadeUniversalMapping(MutableMapping):
+	class FacadeUniversalMapping(Signal, MutableMapping):
 		def __init__(self, engine: AbstractEngine):
+			super().__init__()
 			assert not isinstance(engine, EngineFacade)
 			self.engine = engine
 			self._patch = {}
 			self._deleted = set()
 
 		def _effective_keys(self):
-			return (
-				self.engine.universal.keys() | self._patch.keys()
-			) - self._deleted
+			if self.engine:
+				return (
+					self.engine.universal.keys() | self._patch.keys()
+				) - self._deleted
+			return self._patch.keys() - self._deleted
 
 		def __iter__(self):
 			yield from self._effective_keys()
@@ -1057,30 +1075,33 @@ class EngineFacade(AbstractEngine):
 
 		def __contains__(self, item):
 			return item not in self._deleted and (
-				item in self._patch or item in self.engine.universal
+				item in self._patch
+				and (not self.engine or item in self.engine.universal)
 			)
 
 		def __getitem__(self, item):
 			if item in self._patch:
 				ret = self._patch[item]
-				if ret is None:
+				if ret is ...:
 					raise KeyError("Universal key deleted", item)
 				return ret
-			elif item in self.engine.universal:
+			elif self.engine and item in self.engine.universal:
 				return self.engine.universal[item]
 			else:
 				raise KeyError("No universal key", item)
 
 		def __setitem__(self, key, value):
 			self._patch[key] = value
-			if value is not None:
+			if value is not ...:
 				self._deleted.discard(key)
+			self.send(self, key=key, value=value)
 
 		def __delitem__(self, key):
 			if key not in self.engine.universal:
 				raise KeyError("No key to delete", key)
-			self._patch[key] = None
+			self._patch[key] = ...
 			self._deleted.add(key)
+			self.send(self, key=key, value=...)
 
 	class FacadeCharacterMapping(Mapping):
 		def __init__(self, engine: "EngineFacade"):
@@ -1150,7 +1171,7 @@ class EngineFacade(AbstractEngine):
 			self._created = cache.db._btt()
 			UnitnessCache.__init__(self, cache.db, "unitness_cache")
 			self.user_cache = EngineFacade.FacadeCache(
-				cache.user_cache, "user_cache"
+				cache.leader_cache, "user_cache"
 			)
 			self._real = cache
 
@@ -1205,8 +1226,19 @@ class EngineFacade(AbstractEngine):
 					real._unitness_cache
 				)
 		else:
-			self._branches_d = {}
+			self._branches_d = {
+				"trunk": (
+					None,
+					0,
+					0,
+					0,
+					0,
+				)
+			}
 			self._turn_end_plan = {}
+			self.branch = "trunk"
+			self.turn = 0
+			self.tick = 0
 
 	def _get_node(
 		self, char: AbstractCharacter | CharName, node: NodeName
@@ -1328,7 +1360,7 @@ class EngineFacade(AbstractEngine):
 							char, node, k, v = tup
 							realchar = realeng.character[char]
 							if node in realchar.node:
-								if k is None:
+								if k is ...:
 									realchar.remove_node(node)
 								elif k == "location":
 									# assume the location really exists, since
@@ -1353,7 +1385,7 @@ class EngineFacade(AbstractEngine):
 								orig in realchar.portal
 								and dest in realchar.portal[orig]
 							):
-								if k is None:
+								if k is ...:
 									realchar.remove_portal(orig, dest)
 								else:
 									realchar.portal[orig][dest][k] = v
