@@ -18,12 +18,16 @@ from abc import ABC, abstractmethod
 from collections import defaultdict
 from collections.abc import MutableMapping
 from itertools import chain
+from typing import TYPE_CHECKING
 
 import networkx
 from networkx.exception import NetworkXError
 
-from .typing import Branch, Key, Tick, Turn
-from .wrap import MutableMappingUnwrapper
+from .types import Branch, Key, Tick, Turn
+from .wrap import MutableMappingUnwrapper, SpecialMapping, wrapval
+
+if TYPE_CHECKING:
+	from .engine import Engine
 
 
 class EntityCollisionError(ValueError):
@@ -42,7 +46,7 @@ def getatt(attribute_name):
 _alleged_receivers = defaultdict(list)
 
 
-class AllegedMapping(MutableMappingUnwrapper, ABC):
+class AllegedMapping(MutableMappingUnwrapper, SpecialMapping, ABC):
 	"""Common amenities for mappings"""
 
 	__slots__ = ()
@@ -56,7 +60,7 @@ class AllegedMapping(MutableMappingUnwrapper, ABC):
 
 class AbstractEntityMapping(AllegedMapping, ABC):
 	__slots__ = ()
-	db: "lisien.engine.Engine"
+	db: "Engine"
 
 	@abstractmethod
 	def _get_cache(
@@ -80,16 +84,12 @@ class AbstractEntityMapping(AllegedMapping, ABC):
 	def _set_cache(self, key, branch, turn, tick, value):
 		raise NotImplementedError
 
-	def _set_cache_now(self, key, value):
-		branch, turn, tick = self.db._nbtt()
-		self._set_cache(key, branch, turn, tick, value)
-
 	def _del_db(self, key, branch, turn, tick):
 		"""Delete a key from the database (not the cache)."""
-		self._set_db(key, branch, turn, tick, None)
+		self._set_db(key, branch, turn, tick, ...)
 
 	def _del_cache(self, key, branch, turn, tick):
-		self._set_cache(key, branch, turn, tick, None)
+		self._set_cache(key, branch, turn, tick, ...)
 
 	def __getitem__(self, key):
 		"""If key is 'graph', return myself as a dict, else get the present
@@ -97,45 +97,16 @@ class AbstractEntityMapping(AllegedMapping, ABC):
 
 		"""
 
-		def wrapval(v):
-			from functools import partial
-
-			from .wrap import DictWrapper, ListWrapper, SetWrapper
-
-			if isinstance(v, list):
-				return ListWrapper(
-					partial(self._get_cache_now, key),
-					partial(self._set_cache_now, key),
-					self,
-					key,
-				)
-			elif isinstance(v, dict):
-				return DictWrapper(
-					partial(self._get_cache_now, key),
-					partial(self._set_cache_now, key),
-					self,
-					key,
-				)
-			elif isinstance(v, set):
-				return SetWrapper(
-					partial(self._get_cache_now, key),
-					partial(self._set_cache_now, key),
-					self,
-					key,
-				)
-			else:
-				return v
-
-		return wrapval(self._get_cache_now(key))
+		return wrapval(self, key, self._get_cache_now(key))
 
 	def __contains__(self, item):
 		return item == "name" or self._cache_contains(item, *self.db._btt())
 
 	def __setitem__(self, key, value):
 		"""Set key=value at the present branch and revision"""
-		if value is None:
+		if value is ...:
 			raise ValueError(
-				"Lisien uses None to indicate that a key's been deleted"
+				"Lisien uses the ellipsis to indicate that a key's been deleted"
 			)
 		branch, turn, tick = self.db._nbtt()
 		try:
@@ -168,15 +139,15 @@ class GraphMapping(AbstractEntityMapping):
 	)
 
 	def __init__(self, graph):
-		super().__init__()
+		super().__init__(graph)
 		self.graph = graph
 		self.db = db = graph.db
 		btt = db._btt
 		graph_val_cache = db._graph_val_cache
 		graphn = graph.name
-		self._iter_stuff = (graph_val_cache.iter_entity_keys, graphn, btt)
+		self._iter_stuff = (graph_val_cache.iter_keys, graphn, btt)
 		self._cache_contains_stuff = (graph_val_cache.contains_key, graphn)
-		self._len_stuff = (graph_val_cache.count_entities, graphn, btt)
+		self._len_stuff = (graph_val_cache.count_keys, graphn, btt)
 		self._get_stuff = (self._get_cache, btt)
 		graph_val_set = db.query.graph_val_set
 		self._set_db_stuff = (graph_val_set, graphn)
@@ -194,8 +165,8 @@ class GraphMapping(AbstractEntityMapping):
 		return contains_key(graphn, key, branch, turn, tick)
 
 	def __len__(self):
-		count_entities, graphn, btt = self._len_stuff
-		return 1 + count_entities(graphn, *btt())
+		count_keys, graphn, btt = self._len_stuff
+		return 1 + count_keys(graphn, *btt())
 
 	def __getitem__(self, item):
 		if item == "name":
@@ -225,7 +196,7 @@ class GraphMapping(AbstractEntityMapping):
 
 	def _del_db(self, key, branch, turn, tick):
 		graph_val_set, graphn = self._del_db_stuff
-		graph_val_set(graphn, key, branch, turn, tick, None)
+		graph_val_set(graphn, key, branch, turn, tick, ...)
 
 	def clear(self):
 		keys = set(self.keys())
@@ -258,7 +229,6 @@ class Node(AbstractEntityMapping):
 		"graph",
 		"name",
 		"db",
-		"__weakref__",
 		"_iter_stuff",
 		"_cache_contains_stuff",
 		"_len_stuff",
@@ -272,20 +242,30 @@ class Node(AbstractEntityMapping):
 
 	def __init__(self, graph, node):
 		"""Store name and graph"""
-		super().__init__()
+		super().__init__(graph)
 		self.graph = graph
 		self.name = node
 		self.db = db = graph.db
 		node_val_cache = db._node_val_cache
 		graphn = graph.name
 		btt = db._btt
-		self._iter_stuff = (node_val_cache.iter_entity_keys, graphn, node, btt)
+		self._iter_stuff = (
+			node_val_cache.iter_keys,
+			graphn,
+			node,
+			btt,
+		)
 		self._cache_contains_stuff = (
 			node_val_cache.contains_key,
 			graphn,
 			node,
 		)
-		self._len_stuff = (node_val_cache.count_entity_keys, graphn, node, btt)
+		self._len_stuff = (
+			node_val_cache.count_keys,
+			graphn,
+			node,
+			btt,
+		)
 		self._get_cache_stuff = (node_val_cache.retrieve, graphn, node)
 		self._set_db_stuff = (db.query.node_val_set, graphn, node)
 		self._set_cache_stuff = (db._node_val_cache.store, graphn, node)
@@ -348,9 +328,7 @@ class Edge(AbstractEntityMapping):
 		"graph",
 		"orig",
 		"dest",
-		"idx",
 		"db",
-		"__weakref__",
 		"_iter_stuff",
 		"_cache_contains_stuff",
 		"_len_stuff",
@@ -361,27 +339,20 @@ class Edge(AbstractEntityMapping):
 
 	set_db_time = set_cache_time = 0
 
-	def __init__(self, graph, orig, dest, idx=0):
-		"""Store the graph, the names of the nodes, and the index.
-
-		For non-multigraphs the index is always 0.
-
-		"""
-		super().__init__()
+	def __init__(self, graph, orig, dest):
+		super().__init__(graph)
 		self.graph = graph
 		self.db = db = graph.db
 		self.orig = orig
 		self.dest = dest
-		self.idx = idx
 		edge_val_cache = db._edge_val_cache
 		graphn = graph.name
 		btt = db._btt
 		self._iter_stuff = (
-			edge_val_cache.iter_entity_keys,
+			edge_val_cache.iter_keys,
 			graphn,
 			orig,
 			dest,
-			idx,
 			btt,
 		)
 		self._cache_contains_stuff = (
@@ -389,14 +360,12 @@ class Edge(AbstractEntityMapping):
 			graphn,
 			orig,
 			dest,
-			idx,
 		)
 		self._len_stuff = (
-			edge_val_cache.count_entity_keys,
+			edge_val_cache.count_keys,
 			graphn,
 			orig,
 			dest,
-			idx,
 			btt,
 		)
 		self._get_cache_stuff = (
@@ -404,10 +373,9 @@ class Edge(AbstractEntityMapping):
 			graphn,
 			orig,
 			dest,
-			idx,
 		)
-		self._set_db_stuff = (db.query.edge_val_set, graphn, orig, dest, idx)
-		self._set_cache_stuff = (edge_val_cache.store, graphn, orig, dest, idx)
+		self._set_db_stuff = (db.query.edge_val_set, graphn, orig, dest)
+		self._set_cache_stuff = (edge_val_cache.store, graphn, orig, dest)
 
 	def __repr__(self):
 		return "<{} in graph {} from {} to {} containing {}>".format(
@@ -422,28 +390,28 @@ class Edge(AbstractEntityMapping):
 		return str(dict(self))
 
 	def __iter__(self):
-		iter_entity_keys, graphn, orig, dest, idx, btt = self._iter_stuff
-		return iter_entity_keys(graphn, orig, dest, idx, *btt())
+		iter_entity_keys, graphn, orig, dest, btt = self._iter_stuff
+		return iter_entity_keys(graphn, orig, dest, *btt())
 
 	def _cache_contains(self, key, branch, turn, tick):
-		contains_key, graphn, orig, dest, idx = self._cache_contains_stuff
-		return contains_key(graphn, orig, dest, idx, key, branch, turn, tick)
+		contains_key, graphn, orig, dest = self._cache_contains_stuff
+		return contains_key(graphn, orig, dest, key, branch, turn, tick)
 
 	def __len__(self):
-		count_entity_keys, graphn, orig, dest, idx, btt = self._len_stuff
-		return count_entity_keys(graphn, orig, dest, idx, *btt())
+		count_entity_keys, graphn, orig, dest, btt = self._len_stuff
+		return count_entity_keys(graphn, orig, dest, *btt())
 
 	def _get_cache(self, key, branch, turn, tick):
-		retrieve, graphn, orig, dest, idx = self._get_cache_stuff
-		return retrieve(graphn, orig, dest, idx, key, branch, turn, tick)
+		retrieve, graphn, orig, dest = self._get_cache_stuff
+		return retrieve(graphn, orig, dest, key, branch, turn, tick)
 
 	def _set_db(self, key, branch, turn, tick, value):
-		edge_val_set, graphn, orig, dest, idx = self._set_db_stuff
-		edge_val_set(graphn, orig, dest, idx, key, branch, turn, tick, value)
+		edge_val_set, graphn, orig, dest = self._set_db_stuff
+		edge_val_set(graphn, orig, dest, key, branch, turn, tick, value)
 
 	def _set_cache(self, key, branch, turn, tick, value):
-		store, graphn, orig, dest, idx = self._set_cache_stuff
-		store(graphn, orig, dest, idx, key, branch, turn, tick, value)
+		store, graphn, orig, dest = self._set_cache_stuff
+		store(graphn, orig, dest, key, branch, turn, tick, value)
 
 
 class GraphNodeMapping(AllegedMapping):
@@ -455,7 +423,7 @@ class GraphNodeMapping(AllegedMapping):
 	"""Alias to ``self.graph.db``"""
 
 	def __init__(self, graph):
-		super().__init__()
+		super().__init__(graph)
 		self.graph = graph
 
 	def __iter__(self):
@@ -542,7 +510,7 @@ class GraphNodeMapping(AllegedMapping):
 
 	def update(self, m, /, **kwargs):
 		for node, value in chain(m.items(), kwargs.items()):
-			if value is None:
+			if value is ...:
 				del self[node]
 			elif node not in self:
 				self[node] = value
@@ -562,7 +530,7 @@ class GraphEdgeMapping(AllegedMapping):
 	"""Alias to ``self.graph.db``"""
 
 	def __init__(self, graph):
-		super().__init__()
+		super().__init__(graph)
 		self.graph = graph
 		self._cache = {}
 
@@ -634,7 +602,7 @@ class AbstractSuccessors(GraphEdgeMapping):
 		if dest not in self:
 			raise KeyError("No edge {}->{}".format(self.orig, dest))
 		orig, dest = self._order_nodes(dest)
-		return self.db._get_edge(self.graph, orig, dest, 0)
+		return self.db._get_edge(self.graph, orig, dest)
 
 	def __setitem__(self, dest, value):
 		"""Set the edge between my orig and the given dest to the given
@@ -653,7 +621,7 @@ class AbstractSuccessors(GraphEdgeMapping):
 			self.graph.name, orig, dest, 0, branch, turn, tick, True
 		)
 		self.db._edges_cache.store(
-			self.graph.name, orig, dest, 0, branch, turn, tick, True
+			self.graph.name, orig, dest, branch, turn, tick, True
 		)
 		e = self[real_dest]
 		e.clear()
@@ -667,7 +635,7 @@ class AbstractSuccessors(GraphEdgeMapping):
 			self.graph.name, orig, dest, 0, branch, turn, tick, False
 		)
 		self.db._edges_cache.store(
-			self.graph.name, orig, dest, 0, branch, turn, tick, None
+			self.graph.name, orig, dest, branch, turn, tick, None
 		)
 
 	def __repr__(self):
@@ -770,7 +738,7 @@ class DiGraphPredecessorsMapping(GraphEdgeMapping):
 		):
 			try:
 				if self.db._edges_cache.retrieve(
-					self.graph.name, orig, dest, 0, *self.db._btt()
+					self.graph.name, orig, dest, *self.db._btt()
 				):
 					return True
 			except KeyError:
@@ -864,7 +832,6 @@ class DiGraphPredecessorsMapping(GraphEdgeMapping):
 					self.graph.name,
 					orig,
 					self.dest,
-					0,
 					branch,
 					turn,
 					tick,
@@ -873,42 +840,17 @@ class DiGraphPredecessorsMapping(GraphEdgeMapping):
 				e = self._make_edge(orig)
 			e.update(value)
 			self.db._edges_cache.store(
-				self.graph.name, orig, self.dest, 0, branch, turn, tick, True
+				self.graph.name, orig, self.dest, branch, turn, tick, True
 			)
 
 		def __delitem__(self, orig):
 			"""Unset the existence of the edge from the given node to mine"""
 			branch, turn, tick = self.db._nbtt()
-			if "Multi" in self.graph.__class__.__name__:
-				for idx in self[orig]:
-					self.db.query.exist_edge(
-						self.graph.name,
-						orig,
-						self.dest,
-						idx,
-						branch,
-						turn,
-						tick,
-						False,
-					)
-					self.db._edges_cache.store(
-						self.graph.name,
-						orig,
-						self.dest,
-						idx,
-						branch,
-						turn,
-						tick,
-						False,
-					)
-					return
-				else:
-					raise KeyError("No edges from {}".format(orig))
 			self.db.query.exist_edge(
-				self.graph.name, orig, self.dest, 0, branch, turn, tick, False
+				self.graph.name, orig, self.dest, branch, turn, tick, False
 			)
 			self.db._edges_cache.store(
-				self.graph.name, orig, self.dest, 0, branch, turn, tick, None
+				self.graph.name, orig, self.dest, branch, turn, tick, None
 			)
 
 
