@@ -78,19 +78,41 @@ the function's name.
 
 """
 
+from __future__ import annotations
+
 from abc import ABC, abstractmethod
 from ast import parse
-from collections.abc import Hashable, MutableMapping, MutableSequence
+from collections.abc import Hashable, Iterable, MutableMapping, MutableSequence
 from functools import cached_property, partial
-from typing import Callable, Optional
+from typing import TYPE_CHECKING, Any, Callable, Iterator, Literal, Optional
 
 from astunparse import unparse
 from blinker import Signal
 
-from .cache import Cache
+from .cache import FuncListCache
+from .types import (
+	ActionFuncName,
+	Branch,
+	Key,
+	PrereqFuncName,
+	RuleBig,
+	RulebookName,
+	RulebookPriority,
+	RuleFunc,
+	RuleFuncName,
+	RuleName,
+	RuleNeighborhood,
+	Tick,
+	TriggerFuncName,
+	Turn,
+	Value,
+)
 from .util import AbstractEngine, dedent_source
 from .xcollections import FunctionStore
 
+if TYPE_CHECKING:
+	from .engine import Engine
+
 
 def roundtrip_dedent(source):
 	"""Reformat some lines of code into what unparse makes."""
@@ -102,17 +124,17 @@ class RuleFuncList(MutableSequence, Signal, ABC):
 
 	__slots__ = ["rule"]
 	_funcstore: FunctionStore
-	_cache: Cache
-	_setter: Callable
+	_cache: FuncListCache
+	_setter: Callable[[RuleName, Branch, Turn, Tick, list[RuleFuncName]], None]
 
-	def __init__(self, rule):
+	def __init__(self, rule: Rule):
 		super().__init__()
 		self.rule = rule
 
 	def __repr__(self):
 		return f"<class 'lisien.rule.{self.__class__.__name__}' [{', '.join(self._get())}]>"
 
-	def _nominate(self, v):
+	def _nominate(self, v: RuleFunc | str | RuleFuncName) -> RuleFuncName:
 		if callable(v):
 			self._funcstore(v)
 			return v.__name__
@@ -122,19 +144,19 @@ class RuleFuncList(MutableSequence, Signal, ABC):
 			)
 		return v
 
-	def _get(self):
+	def _get(self) -> list[RuleFuncName]:
 		return self._cache.retrieve(self.rule.name, *self.rule.engine._btt())
 
-	def _set(self, v):
+	def _set(self, v: list[RuleFuncName]) -> None:
 		branch, turn, tick = self.rule.engine._nbtt()
-		self._cache.store(self.rule.name, branch, turn, tick, v)
+		self._cache.store(Key(self.rule.name), branch, turn, tick, v)
 		self._setter(self.rule.name, branch, turn, tick, v)
 
 	def __iter__(self):
 		for funcname in self._get():
 			yield getattr(self._funcstore, funcname)
 
-	def __contains__(self, item):
+	def __contains__(self, item: RuleFunc | str | RuleFuncName):
 		if hasattr(item, "__name__"):
 			item = item.__name__
 		return item in self._get()
@@ -142,29 +164,29 @@ class RuleFuncList(MutableSequence, Signal, ABC):
 	def __len__(self):
 		return len(self._get())
 
-	def __getitem__(self, i):
+	def __getitem__(self, i: int):
 		return getattr(self._funcstore, self._get()[i])
 
-	def __setitem__(self, i, v):
+	def __setitem__(self, i: int, v: RuleFunc | str | RuleFuncName):
 		v = self._nominate(v)
 		l = list(self._get())
 		l[i] = v
 		self._set(list(l))
 		self.send(self)
 
-	def __delitem__(self, i):
+	def __delitem__(self, i: int):
 		l = list(self._get())
 		del l[i]
 		self._set(list(l))
 		self.send(self)
 
-	def insert(self, i, v):
+	def insert(self, i: int, v: RuleFunc | str | RuleFuncName) -> None:
 		l = list(self._get())
 		l.insert(i, self._nominate(v))
 		self._set(list(l))
 		self.send(self)
 
-	def append(self, v):
+	def append(self, v: RuleFunc | str | RuleFuncName) -> None:
 		try:
 			old = self._get()
 		except KeyError:
@@ -172,7 +194,12 @@ class RuleFuncList(MutableSequence, Signal, ABC):
 		self._set(old + [self._nominate(v)])
 		self.send(self)
 
-	def index(self, x, start=0, end=None):
+	def index(
+		self,
+		x: RuleFunc | str | RuleFuncName,
+		start: int = 0,
+		end: int | None = None,
+	):
 		if not callable(x):
 			x = getattr(self._funcstore, x)
 		return super().index(x, start, end)
@@ -182,15 +209,17 @@ class TriggerList(RuleFuncList):
 	"""A list of trigger functions for rules"""
 
 	@cached_property
-	def _funcstore(self):
+	def _funcstore(self) -> FunctionStore:
 		return self.rule.engine.trigger
 
 	@cached_property
-	def _cache(self):
+	def _cache(self) -> InitializedEntitylessCache:
 		return self.rule.engine._triggers_cache
 
 	@cached_property
-	def _setter(self):
+	def _setter(
+		self,
+	) -> Callable[[RuleName, Branch, Turn, Tick, list[TriggerFuncName]], None]:
 		return self.rule.engine.query.set_rule_triggers
 
 
@@ -198,15 +227,17 @@ class PrereqList(RuleFuncList):
 	"""A list of prereq functions for rules"""
 
 	@cached_property
-	def _funcstore(self):
+	def _funcstore(self) -> FunctionStore:
 		return self.rule.engine.prereq
 
 	@cached_property
-	def _cache(self):
+	def _cache(self) -> InitializedEntitylessCache:
 		return self.rule.engine._prereqs_cache
 
 	@cached_property
-	def _setter(self):
+	def _setter(
+		self,
+	) -> Callable[[RuleName, Branch, Turn, Tick, list[PrereqFuncName]], None]:
 		return self.rule.engine.query.set_rule_prereqs
 
 
@@ -214,15 +245,17 @@ class ActionList(RuleFuncList):
 	"""A list of action functions for rules"""
 
 	@cached_property
-	def _funcstore(self):
+	def _funcstore(self) -> FunctionStore:
 		return self.rule.engine.action
 
 	@cached_property
-	def _cache(self):
+	def _cache(self) -> InitializedEntitylessCache:
 		return self.rule.engine._actions_cache
 
 	@cached_property
-	def _setter(self):
+	def _setter(
+		self,
+	) -> Callable[[RuleName, Branch, Turn, Tick, list[ActionFuncName]], None]:
 		return self.rule.engine.query.set_rule_actions
 
 
@@ -238,12 +271,12 @@ class RuleFuncListDescriptor:
 	def flid(self):
 		return "_funclist" + str(id(self))
 
-	def __get__(self, obj, type=None):
+	def __get__(self, obj: Rule, type=None) -> RuleFuncList:
 		if not hasattr(obj, self.flid):
 			setattr(obj, self.flid, self.cls(obj))
 		return getattr(obj, self.flid)
 
-	def __set__(self, obj, value):
+	def __set__(self, obj: Rule, value: list[RuleFunc | str | RuleFuncName]):
 		if not hasattr(obj, self.flid):
 			setattr(obj, self.flid, self.cls(obj))
 		flist = getattr(obj, self.flid)
@@ -273,50 +306,61 @@ class Rule:
 
 	"""
 
-	triggers = RuleFuncListDescriptor(TriggerList)
-	prereqs = RuleFuncListDescriptor(PrereqList)
-	actions = RuleFuncListDescriptor(ActionList)
+	triggers: TriggerList = RuleFuncListDescriptor(TriggerList)
+	prereqs: PrereqList = RuleFuncListDescriptor(PrereqList)
+	actions: ActionList = RuleFuncListDescriptor(ActionList)
+	name: RuleName
 
 	@property
-	def neighborhood(self):
+	def neighborhood(self) -> RuleNeighborhood:
 		try:
-			return self.engine._neighborhoods_cache.retrieve(
+			ret = self.engine._neighborhoods_cache.retrieve(
 				self.name, *self.engine._btt()
 			)
 		except KeyError:
 			return None
+		return ret
 
 	@neighborhood.setter
-	def neighborhood(self, neighbors: int):
+	def neighborhood(self, neighbors: RuleNeighborhood):
+		if neighbors is not None and not isinstance(neighbors, int):
+			raise TypeError("Not an int", neighbors)
+		if neighbors < 0:
+			raise ValueError("Can't be negative", neighbors)
 		btt = self.engine._nbtt()
-		self.engine._neighborhoods_cache.store(self.name, *btt, neighbors)
+		self.engine._neighborhoods_cache.store(
+			Key(self.name), *btt, Value(neighbors)
+		)
 		self.engine.query.set_rule_neighborhood(self.name, *btt, neighbors)
 
 	@property
-	def big(self):
+	def big(self) -> RuleBig:
 		try:
 			return self.engine._rule_bigness_cache.retrieve(
 				self.name, *self.engine._btt()
 			)
 		except KeyError:
-			return False
+			return RuleBig(False)
 
 	@big.setter
-	def big(self, big: bool):
+	def big(self, big: bool | RuleBig):
 		btt = self.engine._nbtt()
-		self.engine._rule_bigness_cache.store(self.name, *btt, big)
+		self.engine._rule_bigness_cache.store(Key(self.name), *btt, Value(big))
 		self.engine.query.set_rule_big(self.name, *btt, big)
 
 	def __init__(
 		self,
-		engine,
-		name,
-		triggers=None,
-		prereqs=None,
-		actions=None,
-		neighborhood=None,
-		big=False,
-		create=True,
+		engine: "Engine",
+		name: RuleName | str,
+		triggers: list[TriggerFuncName | Callable[[Any], bool] | str]
+		| None = None,
+		prereqs: list[PrereqFuncName | Callable[[Any], bool] | str]
+		| None = None,
+		actions: list[ActionFuncName | Callable[[Any], bool] | str]
+		| None = None,
+		neighborhood: RuleNeighborhood = None,
+		big: RuleBig = RuleBig(False),
+		create: bool = True,
 	):
 		"""Store the engine and my name, make myself a record in the database
 		if needed, and instantiate one FunList each for my triggers,
@@ -324,7 +368,10 @@ class Rule:
 
 		"""
 		self.engine = engine
-		self.name = self.__name__ = name
+		self.name = self.__name__ = RuleName(name)
+		# It'd be more correct to subclass InitializedEntitylessCache and
+		# give all its methods more specific signatures that take RuleName.
+		name = Key(name)
 		if create:
 			branch, turn, tick = engine._nbtt()
 			if (
@@ -458,10 +505,14 @@ class Rule:
 					branch, turn, tick, bigkf
 				)
 
-	def __eq__(self, other):
-		return hasattr(other, "name") and self.name == other.name
+	def __eq__(self, other: Rule):
+		return isinstance(other, Rule) and self.name == other.name
 
-	def _fun_names_iter(self, functyp, val):
+	def _fun_names_iter(
+		self,
+		functyp: Literal["trigger", "prereq", "action"],
+		val: Iterable[RuleFunc | str | RuleFuncName],
+	):
 		"""Iterate over the names of the functions in ``val``,
 		adding them to ``funcstore`` if they are missing;
 		or if the items in ``val`` are already the names of functions
@@ -482,39 +533,30 @@ class Rule:
 				yield v
 
 	def __repr__(self):
-		return "Rule({})".format(self.name)
+		return f"<Rule({repr(self.name)})>"
 
-	def trigger(self, fun):
+	def trigger(
+		self, fun: RuleFunc | str | TriggerFuncName
+	) -> RuleFunc | str | TriggerFuncName:
 		"""Decorator to append the function to my triggers list."""
 		self.triggers.append(fun)
 		return fun
 
-	def prereq(self, fun):
+	def prereq(
+		self, fun: RuleFunc | str | PrereqFuncName
+	) -> RuleFunc | str | PrereqFuncName:
 		"""Decorator to append the function to my prereqs list."""
 		self.prereqs.append(fun)
 		return fun
 
-	def action(self, fun):
+	def action(
+		self, fun: RuleFunc | str | ActionFuncName
+	) -> RuleFunc | str | ActionFuncName:
 		"""Decorator to append the function to my actions list."""
 		self.actions.append(fun)
 		return fun
 
-	def duplicate(self, newname):
-		"""Return a new rule that's just like this one, but under a new
-		name.
-
-		"""
-		if self.engine.rule.query.haverule(newname):
-			raise KeyError("Already have a rule called {}".format(newname))
-		return Rule(
-			self.engine,
-			newname,
-			list(self.triggers),
-			list(self.prereqs),
-			list(self.actions),
-		)
-
-	def always(self):
+	def always(self) -> None:
 		"""Arrange to be triggered every turn"""
 		self.triggers = [self.engine.trigger.truth]
 
@@ -522,40 +564,55 @@ class Rule:
 class RuleBook(MutableSequence, Signal):
 	"""A list of rules to be followed for some Character, or a part of it"""
 
-	def _get_cache(self, branch, turn, tick):
+	def _get_cache(
+		self, branch: Branch, turn: Turn, tick: Tick
+	) -> tuple[list[RuleName], RulebookPriority]:
+		rules: list[RuleName] = []
+		prio = RulebookPriority(0.0)
 		try:
 			rules, prio = self.engine._rulebooks_cache.retrieve(
 				self.name, branch, turn, tick
 			)
-			return list(rules), prio
 		except KeyError:
-			return [], 0.0
+			pass
+		return list(rules), prio
 
-	def _set_cache(self, branch, turn, tick, v):
-		self.engine._rulebooks_cache.store(self.name, branch, turn, tick, v)
+	def _set_cache(
+		self,
+		branch: Branch,
+		turn: Turn,
+		tick: Tick,
+		v: tuple[list[RuleName], RulebookPriority],
+	) -> None:
+		with self.engine._rulebooks_cache.overwriting():
+			self.engine._rulebooks_cache.store(
+				self.name, branch, turn, tick, v
+			)
 
-	def __init__(self, engine: "lisien.Engine", name: str):
+	def __init__(self, engine: "Engine", name: RulebookName):
 		super().__init__()
 		self.engine = engine
 		self.name = name
 
 	@property
-	def priority(self):
+	def priority(self) -> RulebookPriority:
 		return self._get_cache(*self.engine._btt())[1]
 
 	@priority.setter
-	def priority(self, v: float):
-		v = float(v)
-		branch, turn, tick = self.engine._btt()
+	def priority(self, v: float | RulebookPriority):
+		v = RulebookPriority(float(v))
+		branch, turn, tick = self.engine._nbtt()
 		cache, _ = self._get_cache(branch, turn, tick)
 		self._set_cache(branch, turn, tick, (cache, v))
 		self.engine.query.set_rulebook(self.name, branch, turn, tick, cache, v)
 
-	def __contains__(self, v):
+	def __contains__(self, v: RuleName | Rule) -> bool:
 		return getattr(v, "name", v) in self._get_cache(*self.engine._btt())[0]
 
-	def __iter__(self):
-		return iter(self._get_cache(*self.engine._btt())[0])
+	def __iter__(self) -> Iterator[Rule]:
+		rule_name: RuleName
+		for rule_name in self._get_cache(*self.engine._btt())[0]:
+			yield self.engine.rule[rule_name]
 
 	def __len__(self):
 		try:
@@ -563,30 +620,34 @@ class RuleBook(MutableSequence, Signal):
 		except KeyError:
 			return 0
 
-	def __getitem__(self, i):
+	def __getitem__(self, i: int) -> Rule:
 		return self.engine.rule[self._get_cache(*self.engine._btt())[0][i]]
 
-	def _coerce_rule(self, v):
+	def _coerce_rule(self, v: Rule | str | RuleName) -> Rule:
 		if isinstance(v, Rule):
 			return v
 		elif isinstance(v, str):
 			return self.engine.rule[v]
 		else:
-			return Rule(self.engine, v)
+			raise TypeError("Invalid rule name", v)
 
-	def __setitem__(self, i, v):
-		v = getattr(v, "name", v)
+	def __setitem__(self, i: int, v: Rule | str | RuleName):
+		v = RuleName(getattr(v, "name", v))
 		if v == "truth":
 			raise ValueError("Illegal rule name")
 		branch, turn, tick = self.engine._nbtt()
+		cache, prio = self._get_cache(branch, turn, tick)
 		try:
-			cache, prio = self._get_cache(branch, turn, tick)
 			cache[i] = v
-		except KeyError:
-			if i != 0:
-				raise IndexError
-			cache = [v]
-			prio = 0.0
+		except IndexError:
+			if i != len(cache):
+				raise
+			if i == 0:
+				cache = [v]
+			else:
+				assert i == len(cache)
+				cache.append(v)
+			prio = RulebookPriority(0.0)
 			self._set_cache(branch, turn, tick, (cache, prio))
 		self.engine.query.set_rulebook(
 			self.name, branch, turn, tick, cache, prio
@@ -594,8 +655,8 @@ class RuleBook(MutableSequence, Signal):
 		self.engine.rulebook.send(self, i=i, v=v)
 		self.send(self, i=i, v=v)
 
-	def insert(self, i, v):
-		v = getattr(v, "name", v)
+	def insert(self, i: int, v: Rule | str | RuleName) -> None:
+		v = RuleName(getattr(v, "name", v))
 		if v == "truth":
 			raise ValueError("Illegal rule name")
 		branch, turn, tick = self.engine._nbtt()
@@ -606,7 +667,7 @@ class RuleBook(MutableSequence, Signal):
 			if i != 0:
 				raise IndexError
 			cache = [v]
-			prio = 0.0
+			prio = RulebookPriority(0.0)
 		self._set_cache(branch, turn, tick, (cache, prio))
 		self.engine.query.set_rulebook(
 			self.name, branch, turn, tick, cache, prio
@@ -614,7 +675,9 @@ class RuleBook(MutableSequence, Signal):
 		self.engine.rulebook.send(self, i=i, v=v)
 		self.send(self, i=i, v=v)
 
-	def index(self, v, start=0, stop=None):
+	def index(
+		self, v: Rule | str | RuleName, start: int = 0, stop: int | None = None
+	) -> int:
 		args = [v, start]
 		if stop is not None:
 			args.append(stop)
@@ -625,7 +688,7 @@ class RuleBook(MutableSequence, Signal):
 				raise ValueError
 		return super().index(*args)
 
-	def __delitem__(self, i):
+	def __delitem__(self, i: int) -> None:
 		branch, turn, tick = self.engine._btt()
 		try:
 			cache, prio = self._get_cache(branch, turn, tick)
@@ -670,7 +733,9 @@ class RuleMapping(MutableMapping, Signal):
 
 	"""
 
-	def __init__(self, engine, rulebook):
+	rulebook: RuleBook
+
+	def __init__(self, engine: "Engine", rulebook: RuleBook | RulebookName):
 		super().__init__()
 		self.engine = engine
 		self._rule_cache = self.engine.rule._cache
@@ -683,31 +748,33 @@ class RuleMapping(MutableMapping, Signal):
 		return "RuleMapping({})".format([k for k in self])
 
 	def __iter__(self):
-		return iter(self.rulebook)
+		return iter(self.rulebook._get_cache(*self.engine._btt())[0])
 
 	def __len__(self):
 		return len(self.rulebook)
 
-	def __contains__(self, k):
+	def __contains__(self, k: Rule | RuleName | str) -> bool:
 		return k in self.rulebook
 
-	def __getitem__(self, k):
+	def __getitem__(self, k: RuleName | str) -> Rule:
 		if k not in self:
 			raise KeyError("Rule '{}' is not in effect".format(k))
 		return self._rule_cache[k]
 
-	def __getattr__(self, k):
+	def __getattr__(self, k: str) -> Rule:
 		if k in self:
 			return self[k]
 		raise AttributeError
 
-	def __setitem__(self, k, v):
+	def __setitem__(
+		self, k: RuleName | str, v: Rule | RuleFunc | RuleName | str
+	):
 		if k == "truth":
 			raise KeyError("Illegal rule name")
 		if isinstance(v, Hashable) and v in self.engine.rule:
 			v = self.engine.rule[v]
-		elif isinstance(v, str) and hasattr(self.engine.function, v):
-			v = getattr(self.engine.function, v)
+		elif isinstance(v, str) and hasattr(self.engine.action, v):
+			v = getattr(self.engine.action, v)
 		if not isinstance(v, Rule) and callable(v):
 			# create a new rule, named k, performing action v
 			self.engine.rule[k] = v
@@ -720,14 +787,14 @@ class RuleMapping(MutableMapping, Signal):
 
 	def __call__(
 		self,
-		v: Optional[callable] = None,
-		name: Optional[str] = None,
+		v: Optional[RuleFunc] = None,
+		name: Optional[RuleName | str] = None,
 		*,
-		neighborhood: Optional[int] = -1,
+		neighborhood: Optional[int] = ...,
 		big: bool = False,
 		always: bool = False,
 	):
-		def wrap(name, v, **kwargs):
+		def wrap(name, v: RuleFunc, **kwargs):
 			name = name if name is not None else v.__name__
 			if name == "truth":
 				raise ValueError("Illegal rule name")
@@ -741,27 +808,27 @@ class RuleMapping(MutableMapping, Signal):
 				r.big = kwargs["big"]
 			return r
 
-		kwargs = {"big": big}
+		kwargs: dict[str, Any] = {"big": big}
 		if always:
 			kwargs["always"] = True
-		if neighborhood != -1:
+		if neighborhood is not ...:
 			kwargs["neighborhood"] = neighborhood
 		if v is None:
 			return partial(wrap, name, **kwargs)
 		return wrap(name, v, **kwargs)
 
-	def __delitem__(self, k):
+	def __delitem__(self, k: RuleName | str):
 		i = self.rulebook.index(k)
 		del self.rulebook[i]
 		self.send(self, key=k, val=None)
 
 	@property
-	def priority(self):
+	def priority(self) -> RulebookPriority:
 		return self.rulebook.priority
 
 	@priority.setter
-	def priority(self, v: float):
-		self.rulebook.priority = v
+	def priority(self, v: RulebookPriority | float):
+		self.rulebook.priority = RulebookPriority(v)
 
 
 class RuleFollower(ABC):
@@ -772,27 +839,28 @@ class RuleFollower(ABC):
 
 	__slots__ = ()
 	engine: AbstractEngine
+	_rulebook: RuleBook
+
+	@cached_property
+	def _rule_mapping(self) -> RuleMapping:
+		return self._get_rule_mapping()
 
 	@property
-	def _rule_mapping(self):
-		if not hasattr(self, "_real_rule_mapping"):
-			self._real_rule_mapping = self._get_rule_mapping()
-		return self._real_rule_mapping
-
-	@property
-	def rule(self, v=None, name=None):
+	def rule(
+		self, v: RuleFunc | None = None, name: RuleName | str | None = None
+	) -> RuleMapping | partial[Rule]:
 		if v is not None:
 			return self._rule_mapping(v, name)
 		return self._rule_mapping
 
 	@property
-	def rulebook(self):
+	def rulebook(self) -> RuleBook:
 		if not hasattr(self, "_rulebook"):
 			self._upd_rulebook()
 		return self._rulebook
 
 	@rulebook.setter
-	def rulebook(self, v):
+	def rulebook(self, v: RuleBook | RulebookName | str):
 		n = v.name if isinstance(v, RuleBook) else v
 		try:
 			if n == self._get_rulebook_name():
@@ -805,10 +873,10 @@ class RuleFollower(ABC):
 	def _upd_rulebook(self):
 		self._rulebook = self._get_rulebook()
 
-	def _get_rulebook(self):
+	def _get_rulebook(self) -> RuleBook:
 		return self.engine.rulebook[self._get_rulebook_name()]
 
-	def rules(self):
+	def rules(self) -> Iterator[Rule]:
 		if not hasattr(self, "engine"):
 			raise AttributeError("Need an engine before I can get rules")
 		return self._rule_mapping.values()
@@ -824,7 +892,7 @@ class RuleFollower(ABC):
 		raise NotImplementedError("_get_rulebook_name")
 
 	@abstractmethod
-	def _set_rulebook_name(self, n):
+	def _set_rulebook_name(self, n: RulebookName | str):
 		"""Tell the database that this is the name of the rulebook to use for
 		me.
 
@@ -835,28 +903,30 @@ class RuleFollower(ABC):
 class AllRuleBooks(MutableMapping, Signal):
 	__slots__ = ["engine", "_cache"]
 
-	def __init__(self, engine):
+	def __init__(self, engine: "Engine"):
 		super().__init__()
 		self.engine = engine
 		self._cache = {}
 
-	def __iter__(self):
+	def __iter__(self) -> Iterator[RulebookName]:
 		return self.engine._rulebooks_cache.iter_entities(*self.engine._btt())
 
 	def __len__(self):
 		return len(list(self))
 
-	def __contains__(self, k):
+	def __contains__(self, k: RulebookName | Key):
 		return self.engine._rulebooks_cache.contains_entity(
 			k, *self.engine._btt()
 		)
 
-	def __getitem__(self, k):
+	def __getitem__(self, k: RulebookName | Key) -> RuleBook:
 		if k not in self._cache:
 			self._cache[k] = RuleBook(self.engine, k)
 		return self._cache[k]
 
-	def __setitem__(self, key, value):
+	def __setitem__(
+		self, key: RulebookName | Key, value: RuleBook | list[Rule | RuleName]
+	):
 		if key not in self._cache:
 			self._cache[key] = RuleBook(self.engine, key)
 		rb = self._cache[key]
@@ -864,7 +934,7 @@ class AllRuleBooks(MutableMapping, Signal):
 			del rb[0]
 		rb.extend(value)
 
-	def __delitem__(self, key):
+	def __delitem__(self, key: RulebookName | Key):
 		self.engine._del_rulebook(key)
 
 
@@ -876,27 +946,29 @@ class AllRules(MutableMapping, Signal):
 
 	"""
 
-	def __init__(self, engine):
+	def __init__(self, engine: "Engine"):
 		super().__init__()
 		self.engine = engine
 
 	@cached_property
-	def _cache(self):
+	def _cache(self) -> dict[RuleName, Rule]:
 		return self.engine._rules_cache
 
-	def __iter__(self):
+	def __iter__(self) -> Iterator[RuleName]:
 		yield from self._cache
 
 	def __len__(self):
 		return len(self._cache)
 
-	def __contains__(self, k):
+	def __contains__(self, k: RuleName | str):
 		return k in self._cache
 
-	def __getitem__(self, k):
+	def __getitem__(self, k: RuleName | str) -> Rule:
 		return self._cache[k]
 
-	def __setitem__(self, k, v):
+	def __setitem__(
+		self, k: RuleName | str, v: Rule | RuleFunc | RuleFuncName | str
+	):
 		# you can use the name of a stored function or rule
 		if isinstance(v, str):
 			if hasattr(self.engine.action, v):
@@ -919,7 +991,7 @@ class AllRules(MutableMapping, Signal):
 			)
 		self.send(self, key=new, rule=v)
 
-	def __delitem__(self, k):
+	def __delitem__(self, k: RuleName | Key):
 		if k not in self:
 			raise KeyError("No such rule")
 		for rulebook in self.engine.rulebooks.values():
@@ -932,13 +1004,13 @@ class AllRules(MutableMapping, Signal):
 
 	def __call__(
 		self,
-		v=None,
-		name=None,
+		v: RuleFunc | RuleFuncName | str | None = None,
+		name: RuleName | str | None = None,
 		*,
-		neighborhood: Optional[int] = -1,
-		always=False,
+		neighborhood: RuleNeighborhood | int | None = ...,
+		always: bool = False,
 	):
-		def r(name, v, **kwargs):
+		def r(name: RuleName | str | None, v: RuleFunc, **kwargs):
 			if name is None:
 				name = v.__name__
 			if name == "truth":
@@ -954,13 +1026,13 @@ class AllRules(MutableMapping, Signal):
 		kwargs = {}
 		if always:
 			kwargs["always"] = True
-		if neighborhood != -1:
+		if neighborhood is not ...:
 			kwargs["neighborhood"] = neighborhood
 		if v is None:
 			return partial(r, name, **kwargs)
 		return r(name, v, **kwargs)
 
-	def new_empty(self, name):
+	def new_empty(self, name: RuleName | str) -> Rule:
 		"""Make a new rule with no actions or anything, and return it."""
 		if name in self:
 			raise KeyError("Already have rule {}".format(name))
