@@ -43,45 +43,60 @@ from operator import (
 from random import Random
 from textwrap import dedent
 from time import monotonic
-from types import FunctionType, MethodType, TracebackType
+from types import FunctionType, MethodType, ModuleType
 from typing import (
-	Any,
+	TYPE_CHECKING,
+	Annotated,
 	Callable,
 	Hashable,
 	Iterable,
+	Iterator,
 	KeysView,
 	MutableMapping,
 	Optional,
 	Sequence,
 	Type,
-	Union,
+	TypeVar,
 )
 
 import msgpack
 import networkx as nx
 import numpy as np
+from annotated_types import Ge, Le
 from blinker import Signal
 from tblib import Traceback
 
 from . import exc
 from .exc import TimeError, WorkerProcessReadOnlyError
-from .graph import DiGraph, Edge, Node
-from .typing import (
+from .graph import DiGraph, Edge, GraphMapping, Node
+from .types import (
 	Branch,
 	CharName,
+	EdgeValDict,
 	EternalKey,
 	NodeName,
+	NodeValDict,
+	RulebookName,
+	RuleName,
 	Stat,
 	Tick,
 	Time,
 	TimeWindow,
 	Turn,
 	UniversalKey,
+	Value,
 )
+from .wrap import SpecialMapping
+
+if TYPE_CHECKING:
+	from .rule import Rule, RuleBook
+	from .xcollections import FunctionStore
 
 TRUE: bytes = msgpack.packb(True)
 FALSE: bytes = msgpack.packb(False)
 NONE: bytes = msgpack.packb(None)
+EMPTY: bytes = msgpack.packb({})
+ELLIPSIS: bytes = b"\xc7\x00{"
 NAME: bytes = msgpack.packb("name")
 NODES: bytes = msgpack.packb("nodes")
 EDGES: bytes = msgpack.packb("edges")
@@ -134,6 +149,7 @@ class MsgpackExtensionType(Enum):
 	place = 0x7E
 	thing = 0x7D
 	portal = 0x7C
+	ellipsis = 0x7B
 	function = 0x7A
 	method = 0x79
 	trigger = 0x78
@@ -187,7 +203,10 @@ def getatt(attribute_name):
 	return property(attrgetter(attribute_name))
 
 
-def singleton_get(s):
+_T = TypeVar("_T")
+
+
+def singleton_get(s: Iterable[_T]) -> _T:
 	"""Take an iterable and return its only item if possible, else None."""
 	it = None
 	for that in s:
@@ -211,17 +230,17 @@ class EntityAccessor(ABC):
 
 	def __init__(
 		self,
-		entity,
-		stat,
-		engine=None,
-		branch=None,
-		turn=None,
-		tick=None,
-		current=False,
-		mungers: list = None,
+		entity: GraphMapping | Node | Edge,
+		stat: Stat,
+		engine: AbstractEngine | None = None,
+		branch: Branch | None = None,
+		turn: Turn | None = None,
+		tick: Tick | None = None,
+		current: bool = False,
+		mungers: list[Callable] | None = None,
 	):
 		if engine is None:
-			engine = entity.engine
+			engine = entity.db
 		if branch is None:
 			branch = engine.branch
 		if turn is None:
@@ -270,7 +289,7 @@ class EntityAccessor(ABC):
 	def __eq__(self, other):
 		return self() == other
 
-	def munge(self, munger):
+	def munge(self, munger: callable):
 		return EntityStatAccessor(
 			self.entity,
 			self.stat,
@@ -310,10 +329,13 @@ class EntityAccessor(ABC):
 		return self.munge(lambda x: x[k])
 
 	@abstractmethod
-	def _get_value_now(self): ...
+	def _get_value_now(self) -> Value: ...
 
 	def __call__(
-		self, branch: Branch = None, turn: Turn = None, tick: Tick = None
+		self,
+		branch: Branch | None = None,
+		turn: Turn | None = None,
+		tick: Tick | None = None,
 	):
 		if self.current:
 			res = self._get_value_now()
@@ -331,7 +353,7 @@ class EntityAccessor(ABC):
 			res = munger(res)
 		return res
 
-	def iter_history(self, beginning, end):
+	def iter_history(self, beginning: Turn, end: Turn) -> Iterator[Value]:
 		"""Iterate over all the values this stat has had in the given window, inclusive."""
 		# It might be useful to do this in a way that doesn't change the
 		# engine's time, perhaps for thread safety
@@ -355,7 +377,7 @@ class EntityAccessor(ABC):
 class UnitsAccessor(EntityAccessor):
 	entity: AbstractCharacter
 
-	def _get_value_now(self):
+	def _get_value_now(self) -> dict[CharName, list[NodeName]]:
 		ret = {}
 		for graph in self.entity.unit:
 			ret[graph] = []
@@ -367,12 +389,12 @@ class UnitsAccessor(EntityAccessor):
 class CharacterStatAccessor(EntityAccessor):
 	entity: AbstractCharacter
 
-	def _get_value_now(self):
+	def _get_value_now(self) -> Value:
 		return self.entity.stat[self.stat]
 
 
 class EntityStatAccessor(EntityAccessor):
-	def _get_value_now(self):
+	def _get_value_now(self) -> Value:
 		return self.entity[self.stat]
 
 
@@ -397,7 +419,7 @@ def _sort_set_key(v):
 class SizedDict(OrderedDict):
 	"""A dictionary that discards old entries when it gets too big."""
 
-	def __init__(self, max_entries=1000):
+	def __init__(self, max_entries: Annotated[int, Ge(0)] = 1000):
 		self._n = max_entries
 		super().__init__()
 
@@ -410,7 +432,7 @@ class SizedDict(OrderedDict):
 _sort_set_memo = SizedDict()
 
 
-def sort_set(s):
+def sort_set(s: Set[_T]) -> list[_T]:
 	"""Return a sorted list of the contents of a set
 
 	This is intended to be used to iterate over world state.
@@ -429,7 +451,15 @@ def sort_set(s):
 	return _sort_set_memo[s].copy()
 
 
-def fake_submit(func, *args, **kwargs):
+class FakeFuture(Future):
+	"""A 'Future' that calls its function immediately and sets the result"""
+
+	def __init__(self, func: callable, *args, **kwargs):
+		super().__init__()
+		self.set_result(func(*args, **kwargs))
+
+
+def fake_submit(func: callable, *args, **kwargs) -> FakeFuture:
 	"""A replacement for `concurrent.futures.Executor.submit` that works in serial
 
 	In testing, you may use, eg.,
@@ -437,15 +467,14 @@ def fake_submit(func, *args, **kwargs):
 	to make normally parallel operations serial.
 
 	"""
-
-	class FakeFuture(Future):
-		def __init__(self, func, *args, **kwargs):
-			super().__init__()
-			self.set_result(func(*args, **kwargs))
-
 	return FakeFuture(func, *args, **kwargs)
 
 
+class AbstractBookmarkMapping(MutableMapping, Callable):
+	@abstractmethod
+	def __call__(self, key: Key) -> None: ...
+
+
 class AbstractEngine(ABC):
 	"""Parent class to the real Engine as well as EngineProxy.
 
@@ -457,23 +486,78 @@ class AbstractEngine(ABC):
 
 	"""
 
-	portal_cls: type
 	thing_cls: type
 	place_cls: type
 	portal_cls: type
 	char_cls: type
 	character: Mapping[CharName, Type[char_cls]]
-	eternal: MutableMapping[EternalKey, Any]
-	universal: MutableMapping[UniversalKey, Any]
+	eternal: MutableMapping[EternalKey, Value]
+	universal: MutableMapping[UniversalKey, Value]
+	rulebook: MutableMapping[RulebookName, "RuleBook"]
+	rule: MutableMapping[RuleName, "Rule"]
+	main_branch: Branch
 	branch: Branch
 	turn: Turn
 	tick: Tick
 	time: Time
+	function: ModuleType | "FunctionStore"
+	method: ModuleType | "FunctionStore"
+	trigger: ModuleType | "FunctionStore"
+	prereq: ModuleType | "FunctionStore"
+	action: ModuleType | "FunctionStore"
+	bookmark: AbstractBookmarkMapping
 	_rando: Random
-	_branches_d: dict[Optional[Branch], TimeWindow]
+	_branches_d: dict[
+		Optional[Branch], tuple[Optional[Branch], Turn, Tick, Turn, Tick]
+	]
 
 	@cached_property
-	def pack(self):
+	def logger(self):
+		if hasattr(self, "_logger"):
+			return self._logger
+		from logging import getLogger
+
+		return getLogger("lisien")
+
+	def log(self, level, msg, *args, **kwargs):
+		self.logger.log(level, msg, *args, **kwargs)
+
+	def debug(self, msg, *args, **kwargs):
+		self.log(10, msg, *args, **kwargs)
+
+	def info(self, msg, *args, **kwargs):
+		self.log(20, msg, *args, **kwargs)
+
+	def warning(self, msg, *args, **kwargs):
+		self.log(30, msg, *args, **kwargs)
+
+	def error(self, msg, *args, **kwargs):
+		self.log(40, msg, *args, **kwargs)
+
+	def critical(self, msg, *args, **kwargs):
+		self.log(50, msg, *args, **kwargs)
+
+	def is_ancestor_of(self, parent: Branch, child: Branch) -> bool:
+		"""Return whether ``child`` is a branch descended from ``parent``
+
+		At any remove.
+
+		"""
+		branches = self.branches()
+		if parent not in branches:
+			raise ValueError("Not a branch", parent)
+		if child not in branches:
+			raise ValueError("Not a branch", child)
+		if parent is None or parent == child or parent == self.main_branch:
+			return True
+		if child == self.main_branch:
+			return False
+		if self.branch_parent(child) == parent:
+			return True
+		return self.is_ancestor_of(parent, self.branch_parent(child))
+
+	@cached_property
+	def pack(self) -> Callable[[Value], bytes]:
 		try:
 			from lise_ormsgpack import packb
 
@@ -481,6 +565,9 @@ class AbstractEngine(ABC):
 		except ImportError:
 			pass
 		handlers = {
+			type(...): lambda _: msgpack.ExtType(
+				MsgpackExtensionType.ellipsis.value, b""
+			),
 			nx.Graph: lambda graf: msgpack.ExtType(
 				MsgpackExtensionType.graph.value,
 				packer(
@@ -545,25 +632,24 @@ class AbstractEngine(ABC):
 				return msgpack.ExtType(
 					MsgpackExtensionType.character.value, packer(obj.name)
 				)
+			elif isinstance(obj, AbstractThing):
+				return msgpack.ExtType(
+					MsgpackExtensionType.thing.value,
+					packer([obj.character.name, obj.name]),
+				)
 			elif isinstance(obj, Node):
-				if hasattr(obj, "location"):
-					return msgpack.ExtType(
-						MsgpackExtensionType.thing.value,
-						packer([obj.character.name, obj.name]),
-					)
-				else:
-					return msgpack.ExtType(
-						MsgpackExtensionType.place.value,
-						packer([obj.character.name, obj.name]),
-					)
+				return msgpack.ExtType(
+					MsgpackExtensionType.place.value,
+					packer([obj.graph.name, obj.name]),
+				)
 			elif isinstance(obj, Edge):
 				return msgpack.ExtType(
 					MsgpackExtensionType.portal.value,
 					packer(
 						[
-							obj.character.name,
-							obj.origin.name,
-							obj.destination.name,
+							obj.graph.name,
+							obj.orig,
+							obj.dest,
 						]
 					),
 				)
@@ -582,10 +668,23 @@ class AbstractEngine(ABC):
 		return packer
 
 	@cached_property
-	def unpack(self):
+	def unpack(
+		self,
+	) -> Callable[
+		[bytes],
+		Value
+		| nx.DiGraph
+		| char_cls
+		| place_cls
+		| thing_cls
+		| portal_cls
+		| Exception
+		| callable,
+	]:
 		char_cls = self.char_cls
 		place_cls = self.place_cls
 		portal_cls = self.portal_cls
+		thing_cls = self.thing_cls
 		excs = {
 			# builtin exceptions
 			"AssertionError": AssertionError,
@@ -635,8 +734,7 @@ class AbstractEngine(ABC):
 			"NetworkXUnfeasible": nx.exception.NetworkXUnfeasible,
 			# lisien exceptions
 			"NonUniqueError": exc.NonUniqueError,
-			"AmbiguousAvatarError": exc.AmbiguousAvatarError,
-			"AmbiguousUserError": exc.AmbiguousUserError,
+			"AmbiguousUserError": exc.AmbiguousLeaderError,
 			"RulesEngineError": exc.RulesEngineError,
 			"RuleError": exc.RuleError,
 			"RedundantRuleError": exc.RedundantRuleError,
@@ -650,7 +748,7 @@ class AbstractEngine(ABC):
 			"WorkerProcessReadOnlyError": exc.WorkerProcessReadOnlyError,
 		}
 
-		def unpack_graph(ext):
+		def unpack_graph(ext: bytes) -> nx.Graph:
 			cls, node, adj, graph = unpacker(ext)
 			blank = {
 				"Graph": nx.Graph,
@@ -663,8 +761,8 @@ class AbstractEngine(ABC):
 			blank.graph = graph
 			return blank
 
-		def unpack_exception(ext):
-			data = unpacker(ext)
+		def unpack_exception(ext: bytes) -> Exception:
+			data: tuple[str, dict | None] = unpacker(ext)
 			if data[0] not in excs:
 				return Exception(*data)
 			ret = excs[data[0]](*data[2:])
@@ -672,57 +770,72 @@ class AbstractEngine(ABC):
 				ret.__traceback__ = Traceback.from_dict(data[1]).to_traceback()
 			return ret
 
-		def unpack_char(ext):
+		def unpack_char(ext: bytes) -> char_cls:
 			charn = unpacker(ext)
 			return char_cls(self, charn, init_rulebooks=False)
 
-		def unpack_place(ext):
+		def unpack_place(ext: bytes) -> place_cls:
 			charn, placen = unpacker(ext)
 			return place_cls(
 				char_cls(self, charn, init_rulebooks=False), placen
 			)
 
-		def unpack_thing(ext):
+		def unpack_thing(ext: bytes) -> thing_cls:
 			charn, thingn = unpacker(ext)
 			# Breaks if the thing hasn't been instantiated yet, not great
 			return self.character[charn].thing[thingn]
 
-		def unpack_portal(ext):
+		def unpack_portal(ext: bytes) -> portal_cls:
 			charn, orign, destn = unpacker(ext)
 			return portal_cls(
 				char_cls(self, charn, init_rulebooks=False), orign, destn
 			)
 
+		def unpack_seq(t: type[_T], ext: bytes) -> _T:
+			unpacked = unpacker(ext)
+			if not isinstance(unpacked, list):
+				raise TypeError("Tried to unpack", type(unpacked), t)
+			return t(unpacked)
+
+		def unpack_func(store: FunctionStore, ext: bytes) -> callable:
+			unpacked = unpacker(ext)
+			if not isinstance(unpacked, str):
+				raise TypeError("Tried to unpack as func", type(unpacked))
+			return getattr(store, unpacked)
+
 		handlers = {
+			MsgpackExtensionType.ellipsis.value: lambda _: ...,
 			MsgpackExtensionType.graph.value: unpack_graph,
 			MsgpackExtensionType.character.value: unpack_char,
 			MsgpackExtensionType.place.value: unpack_place,
 			MsgpackExtensionType.thing.value: unpack_thing,
 			MsgpackExtensionType.portal.value: unpack_portal,
-			MsgpackExtensionType.tuple.value: lambda ext: tuple(unpacker(ext)),
-			MsgpackExtensionType.frozenset.value: lambda ext: frozenset(
-				unpacker(ext)
+			MsgpackExtensionType.tuple.value: partial(unpack_seq, tuple),
+			MsgpackExtensionType.frozenset.value: partial(
+				unpack_seq, frozenset
 			),
-			MsgpackExtensionType.set.value: lambda ext: set(unpacker(ext)),
-			MsgpackExtensionType.function.value: lambda ext: getattr(
-				self.function, unpacker(ext)
+			MsgpackExtensionType.set.value: partial(unpack_seq, set),
+			MsgpackExtensionType.function.value: partial(
+				unpack_func, self.function
 			),
-			MsgpackExtensionType.method.value: lambda ext: getattr(
-				self.method, unpacker(ext)
+			MsgpackExtensionType.method.value: partial(
+				unpack_func, self.method
 			),
-			MsgpackExtensionType.trigger.value: lambda ext: getattr(
-				self.trigger, unpacker(ext)
+			MsgpackExtensionType.trigger.value: partial(
+				unpack_func, self.trigger
 			),
-			MsgpackExtensionType.prereq.value: lambda ext: getattr(
-				self.prereq, unpacker(ext)
+			MsgpackExtensionType.prereq.value: partial(
+				unpack_func, self.prereq
 			),
-			MsgpackExtensionType.action.value: lambda ext: getattr(
-				self.action, unpacker(ext)
+			MsgpackExtensionType.action.value: partial(
+				unpack_func, self.action
 			),
 			MsgpackExtensionType.exception.value: unpack_exception,
 		}
 
-		def unpack_handler(code, data):
+		def unpack_handler(
+			code: MsgpackExtensionType, data: bytes
+		) -> Value | Exception | msgpack.ExtType:
 			if code in handlers:
 				return handlers[code](data)
 			return msgpack.ExtType(code, data)
@@ -745,6 +858,9 @@ class AbstractEngine(ABC):
 		self, char: AbstractCharacter | CharName, node: NodeName
 	) -> Node: ...
 
+	@abstractmethod
+	def _btt(self) -> tuple[Branch, Turn, Tick]: ...
+
 	def branches(self) -> KeysView:
 		return self._branches_d.keys()
 
@@ -794,31 +910,33 @@ class AbstractEngine(ABC):
 		return self._branch_end(branch)[1]
 
 	@abstractmethod
-	def turn_end(self, branch: Branch = None, turn: Turn = None) -> Tick: ...
+	def turn_end(
+		self, branch: Branch | None = None, turn: Turn | None = None
+	) -> Tick: ...
 
 	@abstractmethod
 	def turn_end_plan(
-		self, branch: Branch = None, turn: Turn = None
+		self, branch: Branch | None = None, turn: Turn | None = None
 	) -> Tick: ...
 
 	@abstractmethod
 	def add_character(
 		self,
 		name: CharName,
-		data: Optional[nx.Graph | DiGraph] = None,
+		data: nx.Graph | DiGraph | None = None,
 		layout: bool = False,
-		node: dict = None,
-		edge: dict = None,
+		node: NodeValDict | None = None,
+		edge: EdgeValDict | None = None,
 		**kwargs,
 	): ...
 
 	def new_character(
 		self,
 		name: CharName,
-		data: Optional[nx.Graph | DiGraph] = None,
+		data: nx.Graph | DiGraph | None = None,
 		layout: bool = False,
-		node: dict = None,
-		edge: dict = None,
+		node: NodeValDict | None = None,
+		edge: EdgeValDict | None = None,
 		**kwargs,
 	):
 		self.add_character(name, data)
@@ -828,11 +946,13 @@ class AbstractEngine(ABC):
 		"""Return True or False with equal probability."""
 		return self.choice((True, False))
 
-	def die_roll(self, d: int) -> int:
+	def die_roll(self, d: Annotated[int, Ge(1)]) -> int:
 		"""Roll a die with ``d`` faces. Return the result."""
 		return self.randint(1, d)
 
-	def dice(self, n: int, d: int) -> Iterable[int]:
+	def dice(
+		self, n: Annotated[int, Ge(1)], d: Annotated[int, Ge(1)]
+	) -> Iterable[int]:
 		"""Roll ``n`` dice with ``d`` faces, and yield the results.
 
 		This is an iterator. You'll get the result of each die in
@@ -844,10 +964,10 @@ class AbstractEngine(ABC):
 
 	def dice_check(
 		self,
-		n: int,
-		d: int,
-		target: int,
-		comparator: str | callable = "<=",
+		n: Annotated[int, Ge(1)],
+		d: Annotated[int, Ge(1)],
+		target: Annotated[int, Ge(0)],
+		comparator: str | Callable[[int, int], bool] = "<=",
 	) -> bool:
 		"""Roll ``n`` dice with ``d`` sides, sum them, and compare
 
@@ -870,18 +990,31 @@ class AbstractEngine(ABC):
 			comparator = comps[comparator]
 		return comparator(sum(self.dice(n, d)), target)
 
-	def percent_chance(self, pct: int) -> bool:
+	def chance(self, f: Annotated[float, Ge(0.0), Le(1.0)]) -> bool:
+		"""Return True or False with a given unit probability
+
+		Supply a float between 0.0 and 1.0 to express the probability--
+		or use `percent_chance`
+
+		"""
+		if f <= 0.0:
+			return False
+		if f >= 1.0:
+			return True
+		return f > self._rando.random()
+
+	def percent_chance(
+		self,
+		pct: Annotated[int, Ge(0), Le(100)]
+		| Annotated[float, Ge(0.0), Le(100.0)],
+	) -> bool:
 		"""Return True or False with a given percentile probability
 
 		Values not between 0 and 100 are treated as though they
 		were 0 or 100, whichever is nearer.
 
 		"""
-		if pct <= 0:
-			return False
-		if pct >= 100:
-			return True
-		return pct > self.randint(0, 99)
+		return self.chance(pct / 100)
 
 	betavariate = get_rando("_rando.betavariate")
 	choice = get_rando("_rando.choice")
@@ -903,28 +1036,6 @@ class AbstractEngine(ABC):
 	weibullvariate = get_rando("_rando.weibullvariate")
 
 
-class SpecialMappingDescriptor:
-	def __init__(self, mapclsname):
-		self.mapps = {}
-		self.mapclsname = mapclsname
-
-	def __get__(self, instance, owner):
-		attname = "_" + self.mapclsname
-		if not hasattr(instance, attname):
-			mappcls = getattr(instance, self.mapclsname)
-			setattr(instance, attname, mappcls(instance))
-		return getattr(instance, attname)
-
-	def __set__(self, instance, value):
-		attname = "_" + self.mapclsname
-		if not hasattr(instance, attname):
-			mappcls = getattr(instance, self.mapclsname)
-			setattr(instance, attname, mappcls(instance))
-		it = getattr(instance, attname)
-		it.clear()
-		it.update(value)
-
-
 class AbstractCharacter(DiGraph):
 	"""The Character API, with all requisite mappings and graph generators.
 
@@ -1105,17 +1216,55 @@ class AbstractCharacter(DiGraph):
 	def __getitem__(self, k: NodeName):
 		return self.adj[k]
 
-	thing = SpecialMappingDescriptor("ThingMapping")
-	place = SpecialMappingDescriptor("PlaceMapping")
-	node = nodes = _node = SpecialMappingDescriptor("ThingPlaceMapping")
-	portal = adj = succ = edge = _adj = _succ = SpecialMappingDescriptor(
-		"PortalSuccessorsMapping"
-	)
-	preportal = pred = _pred = SpecialMappingDescriptor(
-		"PortalPredecessorsMapping"
-	)
-	unit = SpecialMappingDescriptor("UnitGraphMapping")
-	stat = getatt("graph")
+	ThingMapping: type[SpecialMapping]
+
+	@cached_property
+	def thing(self) -> ThingMapping:
+		return self.ThingMapping(self)
+
+	PlaceMapping: type[SpecialMapping]
+
+	@cached_property
+	def place(self) -> PlaceMapping:
+		return self.PlaceMapping(self)
+
+	ThingPlaceMapping: type[SpecialMapping]
+
+	@cached_property
+	def _node(self) -> ThingPlaceMapping:
+		return self.ThingPlaceMapping(self)
+
+	node: ThingPlaceMapping = getatt("_node")
+	nodes: ThingPlaceMapping = getatt("_node")
+
+	PortalSuccessorsMapping: type[SpecialMapping]
+
+	@cached_property
+	def _succ(self) -> PortalSuccessorsMapping:
+		return self.PortalSuccessorsMapping(self)
+
+	portal: PortalSuccessorsMapping = getatt("_succ")
+	adj: PortalSuccessorsMapping = getatt("_succ")
+	succ: PortalSuccessorsMapping = getatt("_succ")
+	edge: PortalSuccessorsMapping = getatt("_succ")
+	_adj: PortalSuccessorsMapping = getatt("_succ")
+
+	PortalPredecessorsMapping: type[SpecialMapping]
+
+	@cached_property
+	def _pred(self) -> PortalPredecessorsMapping:
+		return self.PortalPredecessorsMapping(self)
+
+	preportal: PortalPredecessorsMapping = getatt("_pred")
+	pred: PortalPredecessorsMapping = getatt("_pred")
+
+	UnitGraphMapping: type[SpecialMapping]
+
+	@cached_property
+	def unit(self) -> UnitGraphMapping:
+		return self.UnitGraphMapping(self)
+
+	stat: GraphMapping = getatt("graph")
 
 	def units(self):
 		for units in self.unit.values():
@@ -1275,9 +1424,10 @@ def normalize_layout(l: dict) -> dict:
 	return dict(zip(ks, zip(map(float, xnorm), map(float, ynorm))))
 
 
-class AbstractThing(ABC):
+class AbstractThing(MutableMapping):
 	character: AbstractCharacter
 	engine: AbstractEngine
+	name: NodeName
 
 	@property
 	def location(self) -> Node:
@@ -1285,7 +1435,10 @@ class AbstractThing(ABC):
 		locn = self["location"]
 		if locn is None:
 			raise AttributeError("Not really a Thing")
-		return self.engine._get_node(self.character, locn)
+		try:
+			return self.engine._get_node(self.character, locn)
+		except KeyError as ex:
+			raise AttributeError("Doesn't really exist") from ex
 
 	@location.setter
 	def location(self, v: Node | NodeName):
@@ -1493,32 +1646,6 @@ def garbage(arg: callable = None, collect: bool = False):
 		return _garbage_dec(arg, collect=collect)
 
 
-def insist(condition: bool, message: str, obj: Optional[Exception] = None):
-	"""If the condition is false, raise ``obj`` or ``RuntimeError``
-
-	Analogous to ``assert``, but doesn't get stripped out in production.
-
-	"""
-	if not condition:
-		if obj is None:
-			exc = RuntimeError(message)
-		elif isinstance(obj, Exception):
-			exc = obj
-		else:
-			exc = RuntimeError(message, obj)
-		try:
-			raise exc
-		except type(exc) as exception:
-			back_frame = sys.exc_info()[2].tb_frame.f_back
-			tb = TracebackType(
-				tb_next=None,
-				tb_frame=back_frame,
-				tb_lasti=back_frame.f_lasti,
-				tb_lineno=back_frame.f_lineno,
-			)
-			raise exception.with_traceback(tb)
-
-
 def world_locked(fn: callable) -> callable:
 	"""Decorator for functions that alter the world state
 
@@ -1557,11 +1684,13 @@ class TimeSignal(Signal, Sequence):
 	def __len__(self):
 		return 3
 
-	def __getitem__(self, i: str | int) -> str | int:
+	def __getitem__(self, i: str | int) -> Branch | Turn | Tick:
 		if i in ("branch", 0):
 			return self.engine.branch
 		if i in ("turn", 1):
 			return self.engine.turn
+		if i in ("tick", 2):
+			return self.engine.tick
 		if isinstance(i, int):
 			raise IndexError(i)
 		else:
@@ -1569,14 +1698,14 @@ class TimeSignal(Signal, Sequence):
 
 	def __setitem__(self, i: str | int, v: str | int) -> None:
 		if i in ("branch", 0):
-			self.engine.branch = v
+			self.engine.branch = Branch(v)
 		elif i in ("turn", 1):
-			self.engine.turn = v
+			self.engine.turn = Turn(v)
+		elif i in ("tick", 2):
+			self.engine.tick = Tick(v)
 		else:
 			exctyp = KeyError if isinstance(i, str) else IndexError
-			raise exctyp(
-				"Can only set branch or turn. Set `Engine.tick` directly if you really want that."
-			)
+			raise exctyp(i)
 
 	def __str__(self):
 		return str(tuple(self))
